/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/crc/crc.c                                                *
 * Created:     2002-03-11 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/crc.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include "crc.h"


#define CRC_BUF_SIZE MISC_BUF_SIZE


static const char    *argv0 = NULL;

static int           par_enc = 0;
static int           par_dec = 0;

static int           par_bin = 0;
static int           par_pipe = 0;
static int           par_quiet = 0;
static unsigned      par_flags = 0;
static int           par_capital = 0;

static unsigned      par_size = 32;
static unsigned long par_poly = CRC_CRC32;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 1, "algorithm", "name", "Set a predefined algorithm" },
	{ 'b', 0, "binary", NULL, "Print CRC as binary [no]" },
	{ 'c', 0, "capital", NULL, "Use capital letters as hex digits [no]" },
	{ 'd', 0, "decode", NULL, "Remove CRC sum from input [no]" },
	{ 'e', 0, "encode", NULL, "Add CRC sum to output [no]" },
	{ 'g', 1, "polynomial", "poly", "Set generator polynomial [crc32]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'm', 0, "reverse-poly", NULL, "Reverse the polynomial [no]" },
	{ 'n', 1, "crc-bits", "int", "Set CRC size [32]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'p', 0, "pipe", NULL, "Pass input through to stdout [no]" },
	{ 'P', 0, "posix", NULL, "Use the posix CRC algorithm [no]" },
	{ 'q', 0, "quiet", NULL, "Quiet operation [no]" },
	{ 'r', 0, "reverse-input", NULL, "Reverse the bit order in the input [no]" },
	{ 'R', 0, "reverse-output", NULL, "Reverse the bit order in the output [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'x', 0, "invert-init", NULL, "Invert the initial remainder [no]" },
	{ 'X', 0, "invert-crc", NULL, "Invert the CRC [no]" },
	{ 256, 0, "crc16", NULL, "Use the default CRC-16 algorithm [no]" },
	{ 257, 0, "crc32", NULL, "Use the default CRC-32 algorithm [no]" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"crc: calculate cyclic redundancy check checksums",
		"usage: crc [options] [input [output]]",
		opts
	);

	printf (
		"\n"
		"recognized algorithms are:\n"
		"  crc12, crc16, ccitt, cdrom, crc32, posix, sfv\n"
	);

	fflush (stdout);
}


static
char crc_get_hex (unsigned char val)
{
	if (val < 10) {
		return ((char) ('0' + val));
	}

	if (par_capital) {
		return ((char) ('A' + val - 10));
	}

	return ((char) ('a' + val - 10));
}

static
void crc_put_hex (FILE *fp, unsigned long val, unsigned bits)
{
	unsigned digit;

	while (bits >= 4) {
		digit = (val >> (bits - 4)) & 0x0f;
		digit = crc_get_hex (digit);
		fputc (digit, fp);
		bits -= 4;
	}
}

static
int do_crc12_enc (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	unsigned short reg;
	crc12_t        crc;

	crc12_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc12_calc (&crc, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc12_done (&crc);

	reg = crc12_get_crc (&crc);

	buf[0] = (unsigned char) ((reg >> 8) & 0xff);
	buf[1] = (unsigned char) (reg & 0xff);

	if (fwrite (buf, 1, 2, out) != 2) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_crc12_dec (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned short reg;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	crc12_t        crc;

	crc12_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE + 2, inp);

	while (n > 2) {
		crc12_calc (&crc, buf, n - 2);

		if (fwrite (buf, 1, n - 2, out) != (n - 2)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		buf[0] = buf[n - 2];
		buf[1] = buf[n - 1];

		n = 2 + fread (buf + 2, 1, CRC_BUF_SIZE, inp);
	}

	reg = buf[0] & 0xff;
	reg = (reg << 8) + (buf[1] & 0xff);

	crc12_done (&crc);

	if (reg != crc12_get_crc (&crc)) {
		if (par_quiet == 0) {
			fprintf (stderr, "%s: crc mismatch\n", argv0);
		}

		return (1);
	}

	return (0);
}

static
int do_crc12 (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 4];
	unsigned short reg;
	crc12_t        crc;

	crc12_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc12_calc (&crc, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc12_done (&crc);

	reg = crc12_get_crc (&crc);

	if (par_bin) {
		buf[0] = (reg >> 8) & 0xff;
		buf[1] = reg & 0xff;
		if (fwrite (buf, 1, 2, out) != 2) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		crc_put_hex (out, reg, 12);
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}


static
int do_crc16_enc (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	unsigned short reg;
	crc16_t        crc;

	crc16_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc16_calc (&crc, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc16_done (&crc);

	reg = crc16_get_crc (&crc);

	buf[0] = (unsigned char) ((reg >> 8) & 0xff);
	buf[1] = (unsigned char) (reg & 0xff);

	if (fwrite (buf, 1, 2, out) != 2) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_crc16_dec (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned short reg;
	unsigned char  buf[CRC_BUF_SIZE + 2];
	crc16_t        crc;

	crc16_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE + 2, inp);

	while (n > 2) {
		crc16_calc (&crc, buf, n - 2);

		if (fwrite (buf, 1, n - 2, out) != (n - 2)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		buf[0] = buf[n - 2];
		buf[1] = buf[n - 1];

		n = 2 + fread (buf + 2, 1, CRC_BUF_SIZE, inp);
	}

	reg = buf[0] & 0xff;
	reg = (reg << 8) + (buf[1] & 0xff);

	crc16_done (&crc);

	if (reg != crc16_get_crc (&crc)) {
		if (par_quiet == 0) {
			fprintf (stderr, "%s: crc mismatch\n", argv0);
		}

		return (1);
	}

	return (0);
}

static
int do_crc16 (FILE *inp, FILE *out)
{
	unsigned long  n;
	unsigned char  buf[CRC_BUF_SIZE + 4];
	unsigned short reg;
	crc16_t        crc;

	crc16_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc16_calc (&crc, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc16_done (&crc);

	reg = crc16_get_crc (&crc);

	if (par_bin) {
		buf[0] = (reg >> 8) & 0xff;
		buf[1] = reg & 0xff;
		if (fwrite (buf, 1, 2, out) != 2) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		crc_put_hex (out, reg, 16);
		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}


static
int do_crc32_enc (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[CRC_BUF_SIZE + 4];
	unsigned long reg;
	crc32_t       crc;

	crc32_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc32_calc (&crc, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc32_done (&crc);

	reg = crc32_get_crc (&crc);

	buf[0] = (unsigned char) ((reg >> 24) & 0xff);
	buf[1] = (unsigned char) ((reg >> 16) & 0xff);
	buf[2] = (unsigned char) ((reg >> 8) & 0xff);
	buf[3] = (unsigned char) (reg & 0xff);

	if (fwrite (buf, 1, 4, out) != 4) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_crc32_dec (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned long reg;
	unsigned char buf[CRC_BUF_SIZE + 4];
	crc32_t       crc;

	crc32_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE + 4, inp);

	while (n > 4) {
		crc32_calc (&crc, buf, n - 4);

		if (fwrite (buf, 1, n - 4, out) != (n - 4)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		buf[0] = buf[n - 4];
		buf[1] = buf[n - 3];
		buf[2] = buf[n - 2];
		buf[3] = buf[n - 1];

		n = 4 + fread (buf + 4, 1, CRC_BUF_SIZE, inp);
	}

	reg = buf[0] & 0xff;
	reg = (reg << 8) + (buf[1] & 0xff);
	reg = (reg << 8) + (buf[2] & 0xff);
	reg = (reg << 8) + (buf[3] & 0xff);

	crc32_done (&crc);

	if (reg != crc32_get_crc (&crc)) {
		if (par_quiet == 0) {
			fprintf (stderr, "%s: crc mismatch\n", argv0);
		}

		return (1);
	}

	return (0);
}

static
int do_crc32 (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[CRC_BUF_SIZE + 8];
	unsigned long reg;
	crc32_t       crc;

	crc32_init (&crc, par_poly, par_flags);

	n = fread (buf, 1, CRC_BUF_SIZE, inp);

	while (n > 0) {
		crc32_calc (&crc, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, CRC_BUF_SIZE, inp);
	}

	crc32_done (&crc);

	reg = crc32_get_crc (&crc);

	if (par_bin) {
		buf[0] = (reg >> 24) & 0xff;
		buf[1] = (reg >> 16) & 0xff;
		buf[2] = (reg >> 8) & 0xff;
		buf[3] = reg & 0xff;
		if (fwrite (buf, 1, 4, out) != 4) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		crc_put_hex (out, reg, 32);

		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

static
int crc_set_poly (const char *str)
{
	if (strcmp (str, "crc12") == 0) {
		par_size = 12;
		par_poly = CRC_CRC12;
	}
	else if (strcmp (str, "crc16") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16;
	}
	else if (strcmp (str, "ccitt") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16_CCITT;
	}
	else if (strcmp (str, "crc16-ccitt") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16_CCITT;
	}
	else if (strcmp (str, "crc32") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
	}
	else if (strcmp (str, "crc32-c") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32_CRC32C;
	}
	else {
		if (str_get_ulng (str, &par_poly)) {
			return (1);
		}
	}

	return (0);
}

static
int crc_set_algorithm (const char *str)
{
	if (strcmp (str, "crc12") == 0) {
		par_size = 12;
		par_poly = CRC_CRC12;
		par_flags = 0;
	}
	else if (strcmp (str, "crc16") == 0) {
		par_size = 16;
		par_poly = CRC_CRC16;
		par_flags = 0;
	}
	else if (strcmp (str, "ccitt")) {
		par_size = 16;
		par_poly = CRC_CRC16_CCITT;
		par_flags = 0;
	}
	else if (strcmp (str, "cdrom") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32_CDROM;
		par_flags = CRC_REVERSE_INP | CRC_REVERSE_OUT;
	}
	else if (strcmp (str, "crc32") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
		par_flags = 0;
	}
	else if (strcmp (str, "posix") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
		par_flags = CRC_POSIX;
	}
	else if (strcmp (str, "sfv") == 0) {
		par_size = 32;
		par_poly = CRC_CRC32;
		par_flags = CRC_REVERSE_INP | CRC_REVERSE_OUT;
		par_flags |= CRC_INV_V0 | CRC_INV_V1;
	}
	else {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("crc");
			return (EXIT_SUCCESS);

		case 'a':
			if (crc_set_algorithm (optarg[0])) {
				fprintf (stderr, "%s: bad algorithm (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'b':
			par_bin = 1;
			break;

		case 'c':
			par_capital = 1;
			break;

		case 'd':
			par_bin = 1;
			par_enc = 0;
			par_dec = 1;
			break;

		case 'e':
			par_bin = 1;
			par_enc = 1;
			par_dec = 0;
			break;

		case 'g':
			if (crc_set_poly (optarg[0])) {
				fprintf (stderr, "%s: bad polynomial (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'l':
			par_flags |= CRC_REVERSE_INP;
			break;

		case 'm':
			par_poly = crc_reverse_poly (par_poly, par_size);
			break;

		case 'n':
			if (str_get_uint (optarg[0], &par_size)) {
				fprintf (stderr, "%s: bad CRC size (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_pipe = 1;
			break;

		case 'P':
			par_flags |= CRC_POSIX;
			break;

		case 'q':
			par_quiet = 1;
			break;

		case 'r':
			par_flags |= CRC_REVERSE_INP;
			break;

		case 'R':
			par_flags |= CRC_REVERSE_OUT;
			break;

		case 'x':
			par_flags |= CRC_INV_V0;
			break;

		case 'X':
			par_flags |= CRC_INV_V1;
			break;

		case 256:
			crc_set_algorithm ("crc16");
			break;

		case 257:
			crc_set_algorithm ("crc32");
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, par_bin ? "wb" : "w");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	if (par_size == 12) {
		if (par_enc) {
			r = do_crc12_enc (inp, out);
		}
		else if (par_dec) {
			r = do_crc12_dec (inp, out);
		}
		else {
			r = do_crc12 (inp, out);
		}
	}
	else if (par_size == 16) {
		if (par_enc) {
			r = do_crc16_enc (inp, out);
		}
		else if (par_dec) {
			r = do_crc16_dec (inp, out);
		}
		else {
			r = do_crc16 (inp, out);
		}
	}
	else if (par_size == 32) {
		if (par_enc) {
			r = do_crc32_enc (inp, out);
		}
		else if (par_dec) {
			r = do_crc32_dec (inp, out);
		}
		else {
			r = do_crc32 (inp, out);
		}
	}
	else {
		fprintf (stderr, "%s: bad CRC size (%u)\n", argv0, par_size);
		r = 1;
	}

	misc_close (out);
	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
