/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/crypt/bfish.c                                            *
 * Created:     2000-08-24 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/strtoint.h>
#include <lib/md5.h>

#ifdef MISC_DOS
#include <time.h>
#include <bios.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif

#ifdef HAVE_TIME_H
#include <time.h>
#endif

#include "bfish.h"
#include "blowfish.h"
#include "ofb64.h"
#include "passwd.h"


#define BFISH_MAGIC 0x42465348UL

#define BFISH_BUF_SIZE MISC_BUF_SIZE


/*
File format:
0       4   Magic ("BFSH")
4       4   Format version (0)
8       8   IV
16      n   encrypted data
n + 16  16  encrypted MD5 sum of IV + unencrypted data
*/

static const char    *argv0 = NULL;
static int           par_enc = 1;
static int           par_verbose = 0;
static unsigned long par_iv = 0;
static int           par_stdin = 0;
static char          *par_prompt = NULL;
static unsigned      par_pwcnt = 0;

static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'e', 0, "encrypt", NULL, "Encrypt [default]" },
	{ 'd', 0, "decrypt", NULL, "Decrypt" },
	{ 'j', 1, "iv", "int", "Set the unique part of the iv (12 bits) [0]" },
	{ 'n', 1, "password-count", "int", "Read the password n times [0]" },
	{ 'p', 1, "prompt", "string", "Set the password prompt string [password:]" },
	{ 's', 0, "stdin", NULL, "Read the password from stdin [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"bfish: encrypt/decrypt files using the blowfish cypher",
		"usage: bfish [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void bfish_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 24) & 0xff;
	tmp[1] = (v >> 16) & 0xff;
	tmp[2] = (v >> 8) & 0xff;
	tmp[3] = v & 0xff;
}

static
unsigned long bfish_get_uint32 (const void *buf, unsigned i)
{
	unsigned long       v;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	v = tmp[0];
	v = (v << 8) | tmp[1];
	v = (v << 8) | tmp[2];
	v = (v << 8) | tmp[3];

	return (v);
}

#if defined(MISC_DOS)

static
int bfish_get_timestamp (unsigned long *sec, unsigned long *usec)
{
	*sec = (unsigned long) time (NULL);

	if (*sec < BFISH_20000101) {
		*sec = BFISH_20000101;
	}
	else {
		*sec -= BFISH_20000101;
	}

	*usec = biostime (0, 0);

	return (0);
}

#elif defined(HAVE_GETTIMEOFDAY)

static
int bfish_get_timestamp (unsigned long *sec, unsigned long *usec)
{
	struct timeval tv;

	if (gettimeofday (&tv, NULL) < 0) {
		return (1);
	}

	*sec = (unsigned long) tv.tv_sec;
	if (*sec < BFISH_20000101) {
		*sec = BFISH_20000101;
	}
	else {
		*sec -= BFISH_20000101;
	}

	*usec = (unsigned long) tv.tv_usec;

	return (0);
}

#elif defined(HAVE_TIME_H)

static
int bfish_get_timestamp (unsigned long *sec, unsigned long *usec)
{
	*sec = (unsigned long) time (NULL);

	if (*sec < BFISH_20000101) {
		*sec = BFISH_20000101;
	}
	else {
		*sec -= BFISH_20000101;
	}

	*usec = 0;

	return (0);
}

#else

static
int bfish_get_timestamp (unsigned long *sec, unsigned long *usec)
{
	*sec = 0;
	*usec = 0;
}

#endif

static
int bfish_get_iv (block64_t *iv)
{
	unsigned long  hi, lo;

	if (bfish_get_timestamp (&lo, &hi)) {
		return (1);
	}

	hi = ((hi & 0xfffffUL) | (par_iv << 20)) & 0xffffffffUL;

	if (par_verbose) {
		fprintf (stderr, "IV1: %08lx%08lx\n", hi, lo);
	}

	blk64_set_32 (iv, 0, hi);
	blk64_set_32 (iv, 1, lo);

	return (0);
}

static
int bfish_get_key (bfish_ctx_t *bfish, unsigned cnt)
{
	unsigned      i, n, m;
	unsigned char buf[256];
	unsigned char tmp[256];

	if (par_prompt == NULL) {
		par_prompt = "bfish password: ";
	}

	if (par_stdin) {
		n = 256;

		if (crypt_get_passwd_fp (stdin, buf, &n)) {
			return (1);
		}
	}
	else {
		i = 0;
		while (i < cnt) {
			if (i == 0) {
				n = 256;

				if (crypt_get_passwd (par_prompt, buf, &n)) {
					return (1);
				}

				i += 1;
			}
			else {
				m = 256;

				if (crypt_get_passwd (par_prompt, tmp, &m)) {
					return (1);
				}

				if ((n != m) || (memcmp (buf, tmp, n) != 0)) {
					fputs ("password mismatch\n", stderr);
					fflush (stderr);
					i = 0;
				}
				else {
					i += 1;
				}
			}
		}
	}

	bfish_set_key (bfish, buf, n);

	return (0);
}

static
int bfish_ofb_enc (FILE *inp, FILE *out)
{
	unsigned long n;
	unsigned char buf[BFISH_BUF_SIZE + 16];
	block64_t     iv;
	md5_t         md5;
	bfish_ctx_t   bfish;
	ofb_ctx_t     ofb;

	if (bfish_get_key (&bfish, (par_pwcnt > 0) ? par_pwcnt : 2)) {
		return (1);
	}

	md5_init (&md5);
	ofb_init (&ofb, &bfish_crypt, &bfish);

	bfish_set_uint32 (buf, 0, BFISH_MAGIC);
	bfish_set_uint32 (buf, 4, 0);

	bfish_get_iv (&iv);
	bfish_crypt (&bfish, &iv, 1);
	blk64_set_buf (&iv, &buf[8]);

	ofb_set_iv_blk (&ofb, &iv);

	md5_calc (&md5, &buf[8], 8);

	if (fwrite (buf, 1, 16, out) != 16) {
		return (1);
	}

	if (par_verbose) {
		fprintf (stderr, "IV2: %08lx%08lx\n", iv.word32[0], iv.word32[1]);
	}

	n = fread (buf, 1, BFISH_BUF_SIZE, inp);

	while (n > 0) {
		md5_calc (&md5, buf, n);
		ofb_crypt (&ofb, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, BFISH_BUF_SIZE, inp);
	}

	md5_done (&md5);
	md5_get_sum (&md5, buf);
	ofb_crypt (&ofb, buf, 16);

	if (fwrite (buf, 1, 16, out) != 16) {
		return (1);
	}

	fflush (out);

	return (0);
}

static
int bfish_ofb_dec (FILE *inp, FILE *out)
{
	unsigned long i, n;
	unsigned long vers;
	unsigned char buf[BFISH_BUF_SIZE + 16];
	md5_t         md5;
	bfish_ctx_t   bfish;
	ofb_ctx_t     ofb;

	if (fread (buf, 1, 16, inp) != 16) {
		return (1);
	}

	if (bfish_get_uint32 (buf, 0) != BFISH_MAGIC) {
		fprintf (stderr, "%s: bad magic\n", argv0);
		return (1);
	}

	vers = bfish_get_uint32 (buf, 4);
	if (vers != 0) {
		fprintf (stderr, "%s: bad version number (%lu)\n", argv0, vers);
		return (1);
	}

	if (bfish_get_key (&bfish, (par_pwcnt > 0) ? par_pwcnt : 1)) {
		return (1);
	}

	md5_init (&md5);
	md5_calc (&md5, &buf[8], 8);

	ofb_init (&ofb, &bfish_crypt, &bfish);
	ofb_set_iv_buf (&ofb, &buf[8]);

	n = fread (buf, 1, BFISH_BUF_SIZE + 16, inp);
	if (n < 16) {
		fprintf (stderr, "%s: input file too short\n", argv0);
		return (1);
	}

	ofb_crypt (&ofb, buf, n);

	n -= 16;

	while (n > 0) {
		md5_calc (&md5, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		for (i = 0; i < 16; i++) {
			buf[i] = buf[n + i];
		}

		n = fread (buf + 16, 1, BFISH_BUF_SIZE, inp);
		ofb_crypt (&ofb, buf + 16, n);
	}

	fflush (out);

	md5_done (&md5);
	md5_get_sum (&md5, buf + 16);

	for (i = 0; i < 16; i++) {
		if (buf[i] != buf[i + 16]) {
			fprintf (stderr, "%s: MD5 mismatch\n", argv0);
			return (1);
		}
	}

	return (0);
}

static
int do_bfish (FILE *inp, FILE *out)
{
	int r;

	if (par_enc) {
		r = bfish_ofb_enc (inp, out);
	}
	else {
		r = bfish_ofb_dec (inp, out);
	}

	return (r);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("bfish");
			return (EXIT_SUCCESS);

		case 'd':
			par_enc = 0;
			break;

		case 'e':
			par_enc = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'j':
			if (str_get_ulng (optarg[0], &par_iv)) {
				fprintf (stderr, "%s: bad value (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'n':
			if (str_get_uint (optarg[0], &par_pwcnt)) {
				fprintf (stderr, "%s: bad value (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_prompt = optarg[0];
			break;

		case 's':
			par_stdin = 1;
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if ((finp == NULL) && par_stdin) {
		fprintf (stderr, "%s: can't read password from stdin\n", argv0);
		return (EXIT_FAILURE);
	}

	if (finp == NULL) {
		inp = stdin;
	}
	else {
		inp = fopen (finp, "rb");
		if (inp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n", argv0, finp);
			return (EXIT_FAILURE);
		}
	}

	if (fout == NULL) {
		out = stdout;
	}
	else {
		out = fopen (fout, "wb");
		if (out == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n", argv0, fout);
			return (EXIT_FAILURE);
		}
	}

	r = do_bfish (inp, out);

	if (out != stdout) {
		fclose (out);
	}

	if (inp != stdin) {
		fclose (inp);
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
