/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/detty/detty.c                                            *
 * Created:     2002-09-10 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include <stdio.h>
#include <stdlib.h>

#include "detty.h"


static const char *argv0 = NULL;

static int      par_trim = 0;
static unsigned par_width = 1024;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'w', 1, "width", "int", "Set terminal width [1024]" },
	{ 'r', 0, "trim", NULL, "Remove spaces and tabs from end of lines [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"detty: remove TTY special characters from a text stream",
		"usage: detty [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
int is_space (unsigned char c)
{
	if ((c == ' ') || (c == '\t')) {
		return (1);
	}

	return (0);
}

static
void trim (unsigned char *buf, unsigned *n)
{
	while ((*n > 0) && is_space (buf[*n - 1])) {
		*n -= 1;
	}
}

static
int write_line (FILE *fp, unsigned char *buf, unsigned n)
{
	if (par_trim) {
		trim (buf, &n);
	}

	if (fwrite (buf, 1, n, fp) != n) {
		return (1);
	}

	fputc ('\n', fp);

	return (0);
}

/* this is far from perfect... */
static
int skip_escape (FILE *fp)
{
	int c;

	while (1) {
		c = fgetc (fp);

		if (c == EOF) {
			return (1);
		}

		if ((c == '[') || (c == ';')) {
			continue;
		}

		if ((c >= '0') && (c <= '9')) {
			continue;
		}

		return (0);
	}

	return (1);
}

static
int detty (FILE *inp, FILE *out)
{
	unsigned      i;
	int           c;
	unsigned char *buf;
	unsigned      buf_i, buf_n;

	buf_i = 0;
	buf_n = 0;

	buf = malloc (par_width);

	if (buf == NULL) {
		return (1);
	}

	while (1) {
		c = fgetc (inp);

		if (c == EOF) {
			break;
		}

		switch (c) {
		case 0x08:
			if (buf_i > 0) {
				for (i = buf_i; i < buf_n; i++) {
					buf[i - 1] = buf[i];
				}

				buf_i -= 1;
				buf_n -= 1;
			}
			break;

		case 0x0a:
			if (write_line (out, buf, buf_n)) {
				c = EOF;
			}

			buf_i = 0;
			buf_n = 0;

			break;

		case 0x0d:
			buf_i = 0;
			break;

		case 0x1b:
			skip_escape (inp);
			break;

		default:
			if (buf_i < par_width) {
				buf[buf_i] = c;
				buf_i += 1;
				if (buf_i > buf_n) {
					buf_n = buf_i;
				}
			}
			break;
		}

		if (c == EOF) {
			break;
		}
	}

	if (buf_n > 0) {
		write_line (out, buf, buf_n);
	}

	free (buf);

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("detty");
			return (EXIT_SUCCESS);

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'r':
			par_trim = 1;
			break;

		case 'w':
			if (str_get_uint (optarg[0], &par_width)) {
				fprintf (stderr, "%s: bad width (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			if (par_width == 0) {
				fprintf (stderr, "%s: width can't be 0\n",
					argv0
				);
				return (EXIT_FAILURE);
			}
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, "w");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	r = detty (inp, out);

	misc_close (out);
	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
