/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/inet/inet.c                                              *
 * Created:     2002-01-20 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/socket.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "inet.h"


static const char     *argv0 = NULL;

static int            par_listen = 1;
static int            par_verbose = 0;

static int            par_stderr = 0;

static char           *par_host = NULL;
static char           *par_addr = NULL;
static unsigned short par_port = 0;

static unsigned       par_cnt = 0;
static unsigned       par_max = 1;
static unsigned long  par_total = 0;

static unsigned       par_delay = 1;

static char           *par_cmd = NULL;
static char           **par_arg = NULL;

static int            par_term = 0;

static int            par_fd;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'c', 1, "command", "[string...]", "Set command to execute" },
	{ 'h', 1, "host", "host[:port]", "Set the host name" },
	{ 'i', 0, "listen", NULL, "Listen to a port [default]" },
	{ 'm', 1, "max", "int", "Set maximum number of connections [1]" },
	{ 'n', 1, "total", "int", "Set total number of connections [0]" },
	{ 'o', 0, "connect", NULL, "Connect to a port" },
	{ 'p', 1, "port", "int", "Set the port" },
	{ 'r', 1, "delay", "int", "Set the retry delay in seconds [1]" },
	{ 's', 0, "stderr", NULL, "Redirect stderr [no]" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"inet: open a tcp connection and execute a program",
		"usage: inet [options] -c command",
		opts
	);

	fflush (stdout);
}

static
void sig_term (int sig)
{
	par_term = 1;
}

static
void sig_pipe (int sig)
{
}

static
void inet_exec (int fd, int dup_stderr)
{
	if (fd != 0) {
		close (0);
		dup2 (fd, 0);
	}

	if (fd != 1) {
		close (1);
		dup2 (fd, 1);
	}

	if ((fd != 2) && dup_stderr) {
		close (2);
		dup2 (fd, 2);
	}

	if (fd > 2) {
		close (fd);
	}

	execvp (par_cmd, par_arg);

	close (0);
	close (1);
	close (2);

	exit (1);
}

/*
 * Fork and exec the command with fd as stdin and stdout and possibly stderr
 */
static
int inet_fork (int fd, int dup_stderr)
{
	pid_t pid;

	if ((par_max > 0) && (par_cnt >= par_max)) {
		return (1);
	}

	pid = fork();

	if (pid < 0) {
		return (1);
	}
	else if (pid > 0) {
		par_cnt += 1;
		return (0);
	}

	close (par_fd);

	inet_exec (fd, dup_stderr);

	exit (1);
}

static
int inet_wait (unsigned cnt)
{
	int r;
	int ret;

	while (1) {
		r = waitpid (-1, &ret, WNOHANG);

		if (r == 0) {
			break;
		}

		par_cnt -= 1;
	}

	while ((cnt > 0) && (par_cnt >= par_max) && (par_term == 0)) {
		r = wait (&ret);
		if (r > 0) {
			par_cnt -= 1;
		}
	}

	return (0);
}

static
int inet_listen (void)
{
	int            fd;
	char           *host;
	unsigned short port;

	par_fd = -1;

	while (par_fd < 0) {
		if (par_term) {
			return (1);
		}

		par_fd = tcp_listen (par_addr, par_port, 16);

		if (par_fd < 0) {
			if (par_delay == 0) {
				fprintf (stderr,
					"%s: can't create socket for port %u\n",
					argv0, par_port
				);
				return (1);
			}

			sleep (par_delay);
		}
	}

	while (par_term == 0) {
		fd = tcp_accept_time (par_fd, 1000);

		if (fd >= 0) {
			if (par_verbose) {
				if (tcp_get_peer (fd, &host, &port) == 0) {
					printf ("%s:%u -> %u\n",
						host,
						(unsigned) port,
						(unsigned) par_port
					);
					fflush (stdout);
					free (host);
				}
			}

			inet_fork (fd, par_stderr);

			close (fd);

			if (par_total > 0) {
				par_total -= 1;

				if (par_total == 0) {
					close (par_fd);

					inet_wait (1);

					return (0);
				}
			}

			inet_wait (par_max);
		}
	}

	close (par_fd);

	return (1);
}

static
int inet_connect (void)
{
	int fd;

	fd = -1;

	while (fd < 0) {
		if (par_term) {
			return (1);
		}

		fd = tcp_connect (par_addr, par_port);

		if (fd < 0) {
			if (par_delay == 0) {
				fprintf (stderr,
					"%s: connecting to %s:%hu failed\n",
					argv0, par_addr, par_port
				);
				return (1);
			}

			sleep (par_delay);
		}
	}

	inet_exec (fd, par_stderr);

	return (1);
}

int main (int argc, char **argv)
{
	int            r;
	char           **optarg;
	unsigned short tmp1;
	unsigned       tmp2;

	argv0 = argv[0];

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		if ((r == 'c') || (r == 0)) {
			par_cmd = optarg[0];
			par_arg = optarg;
			break;
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("inet");
			return (EXIT_SUCCESS);

		case 'h':
			par_host = tcp_addr_get_host (optarg[0]);

			if (tcp_addr_get_port (optarg[0], &tmp1) == 0) {
				par_port = tmp1;
			}
			break;

		case 'i':
			par_listen = 1;
			break;

		case 'm':
			if (str_get_uint (optarg[0], &par_max)) {
				fprintf (stderr,
					"%s: bad connection count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'n':
			if (str_get_ulng (optarg[0], &par_total)) {
				fprintf (stderr,
					"%s: bad connection count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			par_listen = 0;
			break;

		case 'p':
			if (str_get_uint (optarg[0], &tmp2)) {
				fprintf (stderr, "%s: bad port number (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_port = tmp2;
			break;

		case 'r':
			if (str_get_uint (optarg[0], &par_delay)) {
				fprintf (stderr, "%s: bad delay (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 's':
			par_stderr = 1;
			break;

		case 'v':
			par_verbose = 1;
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_cmd == NULL) {
		fprintf (stderr, "%s: need a command\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_host != NULL) {
		par_addr = tcp_host_resolve (par_host);
	}

	signal (SIGINT, &sig_term);
	signal (SIGPIPE, &sig_pipe);

	if (par_listen) {
		r = inet_listen();
	}
	else {
		r = inet_connect();
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
