/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/file.c                                               *
 * Created:     2002-02-05 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/file.h>


#ifdef MISC_DOS

#include <io.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

typedef int ssize_t;

#define MISC_OPEN_FLAGS_OUT (O_WRONLY | O_CREAT | O_TRUNC | O_BINARY)
#define MISC_OPEN_FLAGS_APP (O_WRONLY | O_CREAT | O_APPEND | O_BINARY)
#define MISC_OPEN_FLAGS_INP (O_RDONLY | O_BINARY)

#define MISC_OPEN_MODE_OUT (S_IFREG | S_IREAD | S_IWRITE)
#define MISC_OPEN_MODE_APP (S_IFREG | S_IREAD | S_IWRITE)
#define MISC_OPEN_MODE_INP (0)

#else

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define MISC_OPEN_FLAGS_OUT (O_WRONLY | O_CREAT | O_TRUNC)
#define MISC_OPEN_FLAGS_APP (O_WRONLY | O_CREAT | O_APPEND)
#define MISC_OPEN_FLAGS_INP (O_RDONLY)

#define MISC_OPEN_MODE_OUT (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)
#define MISC_OPEN_MODE_APP (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)
#define MISC_OPEN_MODE_INP (0)

#endif


void file_init (file_t *f)
{
	f->mode = MISC_MODE_NONE;

	f->name = NULL;
	f->fd = -1;
	f->close = 0;

	f->buf_i = 0;
	f->buf_n = 0;
}

void file_set_name (file_t *f, const char *str)
{
	file_init (f);
	f->name = strdup (str);
	f->close = 0;
}

void file_set_fd (file_t *f, int fd, int close)
{
	file_init (f);
	f->fd = fd;
	f->close = close;
}

void file_set_stdout (file_t *f)
{
	file_init (f);
	f->fd = STDOUT_FILENO;
	f->close = 0;
}

void file_set_stdin (file_t *f)
{
	file_init (f);
	f->fd = STDIN_FILENO;
	f->close = 0;
}

int file_open_out (file_t *f)
{
	f->mode = MISC_MODE_OUT;

	if (f->name == NULL) {
		if (f->fd < 0) {
			return (1);
		}
		else {
			return (0);
		}
	}

	f->fd = open (f->name, MISC_OPEN_FLAGS_OUT, MISC_OPEN_MODE_OUT);

	if (f->fd < 0) {
		return (1);
	}

	f->close = 1;

	return (0);
}

int file_open_app (file_t *f)
{
	f->mode = MISC_MODE_APP;

	if (f->name == NULL) {
		if (f->fd < 0) {
			return (1);
		}
		else {
			return (0);
		}
	}

	f->fd = open (f->name, MISC_OPEN_FLAGS_APP, MISC_OPEN_MODE_APP);

	if (f->fd < 0) {
		return (1);
	}

	f->close = 1;

	return (0);
}

int file_open_inp (file_t *f)
{
	f->mode = MISC_MODE_INP;

	if (f->name == NULL) {
		if (f->fd < 0) {
			return (1);
		}
		else {
			return (0);
		}
	}

	f->fd = open (f->name, MISC_OPEN_FLAGS_INP, MISC_OPEN_MODE_INP);

	if (f->fd < 0) {
		return (1);
	}

	f->close = 1;

	return (0);
}

int file_flush (file_t *f)
{
	size_t  n;
	ssize_t r;

	n = (size_t) (f->buf_n - f->buf_i);

	while (n > 0) {
		r = write (f->fd, f->buf + f->buf_i, n);
		if (r < 0) {
			return (1);
		}

		f->buf_i += (unsigned) r;
		n -= (size_t) r;
	}

	f->buf_i = 0;
	f->buf_n = 0;

	return (0);
}

void file_close (file_t *f)
{
	if ((f->mode == MISC_MODE_OUT) || (f->mode == MISC_MODE_APP)) {
		file_flush (f);
	}

	if (f->close) {
		close (f->fd);
	}

	if (f->name != NULL) {
		free (f->name);
		f->name = NULL;
	}

	f->mode = MISC_MODE_NONE;
}

unsigned long file_write (file_t *f, const void *buf, unsigned long n)
{
	ssize_t r;

	if (f->buf_i < f->buf_n) {
		if (file_flush (f)) {
			return (0);
		}
	}

	r = write (f->fd, buf, (size_t) n);
	if (r < 0) {
		return (0);
	}

	return ((unsigned long) r);
}

unsigned long file_write_block (file_t *f, const void *buf, unsigned long n)
{
	unsigned long i, r;
	unsigned char *tmp;

	tmp = (unsigned char *) buf;

	i = 0;

	while (n > 0) {
		r = file_write (f, tmp + i, n);
		if (r == 0) {
			return (i);
		}

		i += r;
		n -= r;
	}

	return (i);
}

unsigned long file_write_string (file_t *f, const char *str)
{
	unsigned long n;

	n = 0;
	while (str[n] != 0) {
		n += 1;
	}

	return (file_write_block (f, str, n));
}

unsigned long file_read (file_t *f, void *buf, unsigned long n)
{
	unsigned long i;
	ssize_t       r;
	unsigned char *tmp;

	tmp = (unsigned char *) buf;

	i = 0;

	while ((i < n) && (f->buf_i < f->buf_n)) {
		tmp[i] = f->buf[f->buf_i];

		i += 1;
		f->buf_i += 1;
	}

	if (i < n) {
		r = read (f->fd, tmp + i, (size_t) (n - i));
		if (r < 0) {
			return (i);
		}

		i += (unsigned long) r;
	}

	return (i);
}

unsigned long file_read_block (file_t *f, void *buf, unsigned long n)
{
	unsigned long i, r;
	unsigned char *tmp;

	tmp = (unsigned char *) buf;

	i = 0;

	while (n > 0) {
		r = file_read (f, tmp + i, n);
		if (r == 0) {
			return (i);
		}

		i += r;
		n -= r;
	}

	return (i);
}

unsigned long file_skip_block (file_t *f, unsigned long n)
{
	unsigned long i, r;
	unsigned char buf[4096];

	i = 0;

	while (n > 0) {
		r = (n > 4096) ? 4096 : n;
		r = file_read (f, buf, r);

		if (r == 0) {
			return (i);
		}

		i += r;
		n -= r;
	}

	return (i);
}

int file_get_char (file_t *f)
{
	int     ret;
	ssize_t r;

	if (f->buf_i >= f->buf_n) {
		r = read (f->fd, f->buf, MISC_FILE_BUF_SIZE);

		f->buf_i = 0;
		f->buf_n = (unsigned) r;
	}

	if (f->buf_i < f->buf_n) {
		ret = f->buf[f->buf_i];
		f->buf_i += 1;
	}
	else {
		ret = EOF;
	}

	return (ret);
}

int file_unget_char (file_t *f, int c)
{
	if (f->buf_i == 0) {
		return (1);
	}

	f->buf_i -= 1;

	f->buf[f->buf_i] = (unsigned char) c;

	return (0);
}

int file_set_char (file_t *f, int c)
{
	if (f->buf_n >= MISC_FILE_BUF_SIZE) {
		file_flush (f);
	}

	f->buf[f->buf_n] = (unsigned char) c;
	f->buf_n += 1;

	return (0);
}

int file_get_byte (file_t *f, unsigned char *c)
{
	ssize_t r;

	if (f->buf_i >= f->buf_n) {
		r = read (f->fd, f->buf, MISC_FILE_BUF_SIZE);

		f->buf_i = 0;
		f->buf_n = (unsigned) r;
	}

	if (f->buf_i >= f->buf_n) {
		return (1);
	}

	*c = f->buf[f->buf_i];
	f->buf_i += 1;

	return (0);
}

int file_set_byte (file_t *f, unsigned char c)
{
	if (f->buf_n >= MISC_FILE_BUF_SIZE) {
		file_flush (f);
	}

	f->buf[f->buf_n] = c;
	f->buf_n += 1;

	return (0);
}

int file_unget_byte (file_t *f, unsigned char c)
{
	if (f->buf_i == 0) {
		return (1);
	}

	f->buf_i -= 1;

	f->buf[f->buf_i] = c;

	return (0);
}
