/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/md5.c                                                *
 * Created:     2001-06-04 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/config.h>
#include <lib/md5.h>

#define md5_f(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define md5_g(x, y, z) (((x) & (z)) | ((y) & ~(z)))
#define md5_h(x, y, z) ((x) ^ (y) ^ (z))
#define md5_i(x, y, z) ((y) ^ ((x) | ~(z)))

#define md5_rol(x, n) \
	(((x) << (n)) | ((unsigned long) ((x) & 0xffffffffUL) >> (32 - (n))))

#define md5_op(op, a, b, c, d, M, s, t) \
	(a) += op ((b), (c), (d)) + (M) + (t); \
	(a) = (b) + md5_rol ((a), (s));


void md5_init (md5_t *md5)
{
	md5->A = 0x67452301UL;
	md5->B = 0xefcdab89UL;
	md5->C = 0x98badcfeUL;
	md5->D = 0x10325476UL;

	md5->cnt[0] = 0;
	md5->cnt[1] = 0;

	md5->buf_i = 0;
}

void md5_copy (md5_t *dst, const md5_t *src)
{
	unsigned i;

	dst->buf_i = src->buf_i;

	for (i = 0; i < src->buf_i; i++) {
		dst->buf[i] = src->buf[i];
	}

	dst->cnt[0] = src->cnt[0];
	dst->cnt[1] = src->cnt[1];

	dst->A = src->A;
	dst->B = src->B;
	dst->C = src->C;
	dst->D = src->D;
}

static inline
void md5_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = v & 0xff;
	tmp[1] = (v >> 8) & 0xff;
	tmp[2] = (v >> 16) & 0xff;
	tmp[3] = (v >> 24) & 0xff;
}

static inline
void md5_add_size (md5_t *md5, unsigned long n)
{
	n &= 0xffffffffUL;

	md5->cnt[0] = (md5->cnt[0] + n) & 0xffffffffUL;

	if (md5->cnt[0] < n) {
		md5->cnt[1] = (md5->cnt[1] + 1) & 0xffffffffUL;
	}
}

static
void md5_blk_frm_buf (unsigned long *dst, const unsigned char *src)
{
	unsigned      i;
	unsigned long v;

	for (i = 0; i < 16; i++) {
		v = src[3];
		v = (v << 8) + src[2];
		v = (v << 8) + src[1];
		v = (v << 8) + src[0];

		dst[i] = v;

		src += 4;
	}
}

static
void md5_calc_block (md5_t *md5, const void *buf)
{
	unsigned long a, b, c, d;
	unsigned long M[16];

	md5_blk_frm_buf (M, buf);

	a = md5->A;
	b = md5->B;
	c = md5->C;
	d = md5->D;

	md5_op (md5_f, a, b, c, d, M[ 0],  7, 0xd76aa478UL);
	md5_op (md5_f, d, a, b, c, M[ 1], 12, 0xe8c7b756UL);
	md5_op (md5_f, c, d, a, b, M[ 2], 17, 0x242070dbUL);
	md5_op (md5_f, b, c, d, a, M[ 3], 22, 0xc1bdceeeUL);
	md5_op (md5_f, a, b, c, d, M[ 4],  7, 0xf57c0fafUL);
	md5_op (md5_f, d, a, b, c, M[ 5], 12, 0x4787c62aUL);
	md5_op (md5_f, c, d, a, b, M[ 6], 17, 0xa8304613UL);
	md5_op (md5_f, b, c, d, a, M[ 7], 22, 0xfd469501UL);
	md5_op (md5_f, a, b, c, d, M[ 8],  7, 0x698098d8UL);
	md5_op (md5_f, d, a, b, c, M[ 9], 12, 0x8b44f7afUL);
	md5_op (md5_f, c, d, a, b, M[10], 17, 0xffff5bb1UL);
	md5_op (md5_f, b, c, d, a, M[11], 22, 0x895cd7beUL);
	md5_op (md5_f, a, b, c, d, M[12],  7, 0x6b901122UL);
	md5_op (md5_f, d, a, b, c, M[13], 12, 0xfd987193UL);
	md5_op (md5_f, c, d, a, b, M[14], 17, 0xa679438eUL);
	md5_op (md5_f, b, c, d, a, M[15], 22, 0x49b40821UL);

	md5_op (md5_g, a, b, c, d, M[ 1],  5, 0xf61e2562UL);
	md5_op (md5_g, d, a, b, c, M[ 6],  9, 0xc040b340UL);
	md5_op (md5_g, c, d, a, b, M[11], 14, 0x265e5a51UL);
	md5_op (md5_g, b, c, d, a, M[ 0], 20, 0xe9b6c7aaUL);
	md5_op (md5_g, a, b, c, d, M[ 5],  5, 0xd62f105dUL);
	md5_op (md5_g, d, a, b, c, M[10],  9, 0x02441453UL);
	md5_op (md5_g, c, d, a, b, M[15], 14, 0xd8a1e681UL);
	md5_op (md5_g, b, c, d, a, M[ 4], 20, 0xe7d3fbc8UL);
	md5_op (md5_g, a, b, c, d, M[ 9],  5, 0x21e1cde6UL);
	md5_op (md5_g, d, a, b, c, M[14],  9, 0xc33707d6UL);
	md5_op (md5_g, c, d, a, b, M[ 3], 14, 0xf4d50d87UL);
	md5_op (md5_g, b, c, d, a, M[ 8], 20, 0x455a14edUL);
	md5_op (md5_g, a, b, c, d, M[13],  5, 0xa9e3e905UL);
	md5_op (md5_g, d, a, b, c, M[ 2],  9, 0xfcefa3f8UL);
	md5_op (md5_g, c, d, a, b, M[ 7], 14, 0x676f02d9UL);
	md5_op (md5_g, b, c, d, a, M[12], 20, 0x8d2a4c8aUL);

	md5_op (md5_h, a, b, c, d, M[ 5],  4, 0xfffa3942UL);
	md5_op (md5_h, d, a, b, c, M[ 8], 11, 0x8771f681UL);
	md5_op (md5_h, c, d, a, b, M[11], 16, 0x6d9d6122UL);
	md5_op (md5_h, b, c, d, a, M[14], 23, 0xfde5380cUL);
	md5_op (md5_h, a, b, c, d, M[ 1],  4, 0xa4beea44UL);
	md5_op (md5_h, d, a, b, c, M[ 4], 11, 0x4bdecfa9UL);
	md5_op (md5_h, c, d, a, b, M[ 7], 16, 0xf6bb4b60UL);
	md5_op (md5_h, b, c, d, a, M[10], 23, 0xbebfbc70UL);
	md5_op (md5_h, a, b, c, d, M[13],  4, 0x289b7ec6UL);
	md5_op (md5_h, d, a, b, c, M[ 0], 11, 0xeaa127faUL);
	md5_op (md5_h, c, d, a, b, M[ 3], 16, 0xd4ef3085UL);
	md5_op (md5_h, b, c, d, a, M[ 6], 23, 0x04881d05UL);
	md5_op (md5_h, a, b, c, d, M[ 9],  4, 0xd9d4d039UL);
	md5_op (md5_h, d, a, b, c, M[12], 11, 0xe6db99e5UL);
	md5_op (md5_h, c, d, a, b, M[15], 16, 0x1fa27cf8UL);
	md5_op (md5_h, b, c, d, a, M[ 2], 23, 0xc4ac5665UL);

	md5_op (md5_i, a, b, c, d, M[ 0],  6, 0xf4292244UL);
	md5_op (md5_i, d, a, b, c, M[ 7], 10, 0x432aff97UL);
	md5_op (md5_i, c, d, a, b, M[14], 15, 0xab9423a7UL);
	md5_op (md5_i, b, c, d, a, M[ 5], 21, 0xfc93a039UL);
	md5_op (md5_i, a, b, c, d, M[12],  6, 0x655b59c3UL);
	md5_op (md5_i, d, a, b, c, M[ 3], 10, 0x8f0ccc92UL);
	md5_op (md5_i, c, d, a, b, M[10], 15, 0xffeff47dUL);
	md5_op (md5_i, b, c, d, a, M[ 1], 21, 0x85845dd1UL);
	md5_op (md5_i, a, b, c, d, M[ 8],  6, 0x6fa87e4fUL);
	md5_op (md5_i, d, a, b, c, M[15], 10, 0xfe2ce6e0UL);
	md5_op (md5_i, c, d, a, b, M[ 6], 15, 0xa3014314UL);
	md5_op (md5_i, b, c, d, a, M[13], 21, 0x4e0811a1UL);
	md5_op (md5_i, a, b, c, d, M[ 4],  6, 0xf7537e82UL);
	md5_op (md5_i, d, a, b, c, M[11], 10, 0xbd3af235UL);
	md5_op (md5_i, c, d, a, b, M[ 2], 15, 0x2ad7d2bbUL);
	md5_op (md5_i, b, c, d, a, M[ 9], 21, 0xeb86d391UL);

	md5->A += a;
	md5->B += b;
	md5->C += c;
	md5->D += d;
}

void md5_calc (md5_t *md5, const void *buf, unsigned long cnt)
{
	unsigned            i, n;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	if (md5->buf_i > 0) {
		while ((md5->buf_i < 64) && (cnt > 0)) {
			md5->buf[md5->buf_i++] = *(tmp++);
			cnt -= 1;
		}

		if (md5->buf_i < 64) {
			return;
		}

		md5_calc_block (md5, md5->buf);

		md5->buf_i = 0;
		md5_add_size (md5, 8 * 64);
	}

	while (cnt >= 64) {
		md5_calc_block (md5, tmp);

		md5_add_size (md5, 8 * 64);

		tmp += 64;
		cnt -= 64;
	}

	n = (unsigned) cnt;
	for (i = 0; i < n; i++) {
		md5->buf[i] = tmp[i];
	}

	md5->buf_i = n;
}

void md5_done (md5_t *md5)
{
	unsigned i;

	md5_add_size (md5, 8 * md5->buf_i);

	md5->buf[md5->buf_i++] = 0x80;

	if (md5->buf_i > 56) {
		for (i = md5->buf_i; i < 64; i++) {
			md5->buf[i] = 0;
		}

		md5_calc_block (md5, md5->buf);

		md5->buf_i = 0;
	}

	for (i = md5->buf_i; i < 56; i++) {
		md5->buf[i] = 0;
	}

	md5_set_uint32 (md5->buf, 56, md5->cnt[0]);
	md5_set_uint32 (md5->buf, 60, md5->cnt[1]);

	md5_calc_block (md5, md5->buf);
}

unsigned long md5_get_sum_i (md5_t *md5, unsigned i)
{
	switch (i) {
		case 0:
			return (md5->A);

		case 1:
			return (md5->B);

		case 2:
			return (md5->C);

		case 3:
			return (md5->D);

		default:
			return (0);
	}

	return (0);
}

void md5_get_sum (md5_t *md5, void *buf)
{
	md5_set_uint32 (buf, 0, md5->A);
	md5_set_uint32 (buf, 4, md5->B);
	md5_set_uint32 (buf, 8, md5->C);
	md5_set_uint32 (buf, 12, md5->D);
}
