/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lib/sha1.c                                               *
 * Created:     2004-04-10 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/config.h>
#include <lib/sha1.h>


#define SHA1_K1 0x5a827999UL
#define SHA1_K2 0x6ed9eba1UL
#define SHA1_K3 0x8f1bbcdcUL
#define SHA1_K4 0xca62c1d6UL

#define SHA1_H0 0x67452301UL
#define SHA1_H1 0xefcdab89UL
#define SHA1_H2 0x98badcfeUL
#define SHA1_H3 0x10325476UL
#define SHA1_H4 0xc3d2e1f0UL

#define sha1_f1(x, y, z) (((x) & (y)) | (~(x) & (z)))
#define sha1_f2(x, y, z) ((x) ^ (y) ^ (z))
#define sha1_f3(x, y, z) (((x) & (y)) | ((x) & (z)) | ((y) & (z)))
#define sha1_f4(x, y, z) ((x) ^ (y) ^ (z))

#define sha1_rol(x, n) \
	(((x) << (n)) | ((unsigned long) ((x) & 0xffffffffUL) >> (32 - (n))))

#define sha1_op(op, a, b, c, d, e, t, W, K) \
		t = sha1_rol (a, 5) + op (b, c, d) + e + W + K; \
		e = d; \
		d = c; \
		c = sha1_rol (b, 30); \
		b = a; \
		a = t;


void sha1_init (sha1_t *sha1)
{
	sha1->buf_i = 0;

	sha1->cnt[0] = 0;
	sha1->cnt[1] = 0;

	sha1->A = SHA1_H0;
	sha1->B = SHA1_H1;
	sha1->C = SHA1_H2;
	sha1->D = SHA1_H3;
	sha1->E = SHA1_H4;
}

void sha1_copy (sha1_t *dst, const sha1_t *src)
{
	unsigned i;

	dst->buf_i = src->buf_i;

	for (i = 0; i < src->buf_i; i++) {
		dst->buf[i] = src->buf[i];
	}

	dst->cnt[0] = src->cnt[0];
	dst->cnt[1] = src->cnt[1];

	dst->A = src->A;
	dst->B = src->B;
	dst->C = src->C;
	dst->D = src->D;
	dst->E = src->E;
}

static inline
void sha1_set_uint32 (void *buf, unsigned i, unsigned long v)
{
	unsigned char *tmp;

	tmp = (unsigned char *) buf + i;

	tmp[0] = (v >> 24) & 0xff;
	tmp[1] = (v >> 16) & 0xff;
	tmp[2] = (v >> 8) & 0xff;
	tmp[3] = v & 0xff;
}

static inline
unsigned long sha1_get_uint32 (const void *buf, unsigned i)
{
	unsigned long       v;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf + i;

	v = tmp[0];
	v = (v << 8) | tmp[1];
	v = (v << 8) | tmp[2];
	v = (v << 8) | tmp[3];

	return (v);
}

static
void sha1_add_size (sha1_t *sha1, unsigned long n)
{
	n &= 0xffffffffUL;

	sha1->cnt[0] = (sha1->cnt[0] + n) & 0xffffffffUL;

	if (sha1->cnt[0] < n) {
		sha1->cnt[1] = (sha1->cnt[1] + 1) & 0xffffffffUL;
	}
}

static
void sha1_blk_frm_buf (unsigned long *W, const void *buf)
{
	unsigned i;

	for (i = 0; i < 16; i++) {
		W[i] = sha1_get_uint32 (buf, 4 * i);
	}

	for (i = 16; i < 80; i++) {
		W[i] = sha1_rol (W[i - 3] ^ W[i - 8] ^ W[i - 14] ^ W[i - 16], 1);
	}
}

static
void sha1_calc_block (sha1_t *sha1, const void *buf)
{
	unsigned long a, b, c, d, e, t;
	unsigned long W[80];

	sha1_blk_frm_buf (W, buf);

	a = sha1->A;
	b = sha1->B;
	c = sha1->C;
	d = sha1->D;
	e = sha1->E;

	sha1_op (sha1_f1, a, b, c, d, e, t, W[0], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[1], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[2], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[3], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[4], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[5], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[6], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[7], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[8], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[9], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[10], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[11], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[12], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[13], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[14], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[15], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[16], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[17], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[18], SHA1_K1);
	sha1_op (sha1_f1, a, b, c, d, e, t, W[19], SHA1_K1);

	sha1_op (sha1_f2, a, b, c, d, e, t, W[20], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[21], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[22], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[23], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[24], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[25], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[26], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[27], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[28], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[29], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[30], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[31], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[32], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[33], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[34], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[35], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[36], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[37], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[38], SHA1_K2);
	sha1_op (sha1_f2, a, b, c, d, e, t, W[39], SHA1_K2);

	sha1_op (sha1_f3, a, b, c, d, e, t, W[40], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[41], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[42], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[43], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[44], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[45], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[46], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[47], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[48], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[49], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[50], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[51], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[52], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[53], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[54], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[55], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[56], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[57], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[58], SHA1_K3);
	sha1_op (sha1_f3, a, b, c, d, e, t, W[59], SHA1_K3);

	sha1_op (sha1_f4, a, b, c, d, e, t, W[60], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[61], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[62], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[63], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[64], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[65], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[66], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[67], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[68], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[69], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[70], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[71], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[72], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[73], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[74], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[75], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[76], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[77], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[78], SHA1_K4);
	sha1_op (sha1_f4, a, b, c, d, e, t, W[79], SHA1_K4);

	sha1->A += a;
	sha1->B += b;
	sha1->C += c;
	sha1->D += d;
	sha1->E += e;
}

void sha1_calc (sha1_t *sha1, const void *buf, unsigned long cnt)
{
	unsigned            i, n;
	const unsigned char *tmp;

	tmp = (const unsigned char *) buf;

	if (sha1->buf_i > 0) {
		while ((sha1->buf_i < 64) && (cnt > 0)) {
			sha1->buf[sha1->buf_i] = *tmp;
			tmp += 1;
			sha1->buf_i += 1;
			cnt -= 1;
		}

		if (sha1->buf_i < 64) {
			return;
		}

		sha1_calc_block (sha1, sha1->buf);

		sha1->buf_i = 0;
		sha1_add_size (sha1, 8 * 64);
	}

	while (cnt >= 64) {
		sha1_calc_block (sha1, tmp);
		sha1_add_size (sha1, 8 * 64);

		tmp += 64;
		cnt -= 64;
	}

	n = (unsigned) cnt;
	for (i = 0; i < n; i++) {
		sha1->buf[i] = tmp[i];
	}

	sha1->buf_i = n;
}

void sha1_done (sha1_t *sha1)
{
	unsigned i;

	sha1_add_size (sha1, 8 * sha1->buf_i);

	sha1->buf[sha1->buf_i++] = 0x80;

	if (sha1->buf_i > 56) {
		for (i = sha1->buf_i; i < 64; i++) {
			sha1->buf[i] = 0;
		}

		sha1_calc_block (sha1, sha1->buf);

		sha1->buf_i = 0;
	}

	for (i = sha1->buf_i; i < 56; i++) {
		sha1->buf[i] = 0;
	}

	sha1_set_uint32 (sha1->buf, 56, sha1->cnt[1]);
	sha1_set_uint32 (sha1->buf, 60, sha1->cnt[0]);

	sha1_calc_block (sha1, sha1->buf);
}

unsigned long sha1_get_sum_i (sha1_t *sha1, unsigned i)
{
	switch (i) {
		case 0:
			return (sha1->A);

		case 1:
			return (sha1->B);

		case 2:
			return (sha1->C);

		case 3:
			return (sha1->D);

		case 4:
			return (sha1->E);

		default:
			return (0);
	}

	return (0);
}

void sha1_get_sum (sha1_t *sha1, void *buf)
{
	sha1_set_uint32 (buf, 0, sha1->A);
	sha1_set_uint32 (buf, 4, sha1->B);
	sha1_set_uint32 (buf, 8, sha1->C);
	sha1_set_uint32 (buf, 12, sha1->D);
	sha1_set_uint32 (buf, 16, sha1->E);
}
