/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/lnconv/lnconv.c                                          *
 * Created:     2000-09-22 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/stresc.h>

#include "lnconv.h"


static const char    *argv0 = NULL;

static int           par_do_lnconv = 0;
static int           par_do_rtrim = 0;

static unsigned char par_eol[LNCONV_STR_MAX];
static unsigned      par_eol_cnt = 0;

static unsigned char par_rtrim[LNCONV_STR_MAX];
static unsigned      par_rtrim_cnt = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'e', 1, "eol", "string", "Set the end of line type [unix]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'r', 0, "rtrim", NULL, "Remove white space from end of lines [no]" },
	{ 'R', 1, "rtrim-chars", "string", "Remove characters in string from end of lines" },
	{ 's', 1, "eol-string", "string", "Set literal end of line string" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"lnconv: convert end of line styles",
		"usage: lnconv [options] [input [output]]",
		opts
	);

	fputs (
		"\nrecognized end of line types are:\n"
		"  lf, unix   (LF)\n"
		"  crlf, dos  (CR+LF)\n"
		"  cr, mac    (CR)\n"
		"  null       (0)\n",
		stdout
	);

	fflush (stdout);
}

static
void line_init (line_t *ln)
{
	ln->i = 0;
	ln->n = 0;
	ln->buf = NULL;

	ln->eol_cnt = 0;
}

static
void line_free (line_t *ln)
{
	free (ln->buf);
}

static
int line_add_char (line_t *ln, unsigned char c)
{
	if (ln->i >= ln->n) {
		unsigned      n;
		unsigned char *tmp;

		n = ln->i;
		while (n <= ln->n) {
			n = 2 * n + 1;
		}

		tmp = realloc (ln->buf, n);
		if (tmp == NULL) {
			return (1);
		}

		ln->buf = tmp;
		ln->n = n;
	}

	ln->buf[ln->i] = c;
	ln->i += 1;

	return (0);
}

static
int line_read (line_t *ln, FILE *fp)
{
	int c;

	ln->i = 0;
	ln->eol_cnt = 0;

	c = fgetc (fp);
	while (c != EOF) {
		switch (c) {
			case 0x0d:
				ln->eol[0] = 0x0d;
				ln->eol_cnt = 1;

				c = fgetc (fp);
				if (c == 0x0a) {
					ln->eol[1] = 0x0a;
					ln->eol_cnt = 2;
				}
				else {
					ungetc (c, fp);
				}
				return (0);

			case 0x0a:
				ln->eol[0] = 0x0a;
				ln->eol_cnt = 1;
				return (0);

			default:
				if (line_add_char (ln, c)) {
					return (1);
				}
				break;
		}

		c = fgetc (fp);
	}

	return (0);
}

static
int line_write (line_t *ln, FILE *fp)
{
	if (ln->i > 0) {
		if (fwrite (ln->buf, 1, ln->i, fp) != ln->i) {
			return (1);
		}
	}

	if (ln->eol_cnt > 0) {
		if (fwrite (ln->eol, 1, ln->eol_cnt, fp) != ln->eol_cnt) {
			return (1);
		}
	}

	return (0);
}

static
int line_set_eol (line_t *ln, unsigned char *buf, unsigned n)
{
	if (n > LNCONV_STR_MAX) {
		return (1);
	}

	memcpy (ln->eol, buf, n);
	ln->eol_cnt = n;

	return (0);
}

static
int char_in_set (unsigned char c, const unsigned char *buf, unsigned n)
{
	while (n > 0) {
		if (*buf == c) {
			return (1);
		}

		buf += 1;
		n -= 1;
	}

	return (0);
}

static
void line_rtrim (line_t *ln, unsigned char *buf, unsigned n)
{
	while (ln->i > 0) {
		if (char_in_set (ln->buf[ln->i - 1], buf, n)) {
			ln->i -= 1;
		}
		else {
			return;
		}
	}
}

static
int line_is_eof (line_t *ln)
{
	if ((ln->i == 0) && (ln->eol_cnt == 0)) {
		return (1);
	}

	return (0);
}

static
int lnconv (FILE *inp, FILE *out)
{
	line_t ln;

	line_init (&ln);

	while (line_read (&ln, inp) == 0) {
		if (line_is_eof (&ln)) {
			break;
		}

		if (par_do_lnconv) {
			line_set_eol (&ln, par_eol, par_eol_cnt);
		}

		if (par_do_rtrim) {
			line_rtrim (&ln, par_rtrim, par_rtrim_cnt);
		}

		if (line_write (&ln, out)) {
			line_free (&ln);
			return (1);
		}
	}

	line_free (&ln);

	return (0);
}

static
int set_eol_style (const char *str)
{
	if ((strcmp (str, "lf") == 0) || (strcmp (str, "unix") == 0)) {
		par_eol[0] = 0x0a;
		par_eol_cnt = 1;
	}
	else if ((strcmp (str, "cr") == 0) || (strcmp (str, "mac") == 0)) {
		par_eol[0] = 0x0d;
		par_eol_cnt = 1;
	}
	else if ((strcmp (str, "crlf") == 0) || (strcmp (str, "dos") == 0)) {
		par_eol[0] = 0x0d;
		par_eol[1] = 0x0a;
		par_eol_cnt = 2;
	}
	else if (strcmp (str, "null") == 0) {
		par_eol[0] = 0x00;
		par_eol_cnt = 1;
	}
	else {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;
	char *finp, *fout;
	FILE *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("lnconv");
			return (EXIT_SUCCESS);

		case 'e':
			if (set_eol_style (optarg[0])) {
				fprintf (stderr,
					"%s: unknown end of line type (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_do_lnconv = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'r':
			par_do_rtrim = 1;
			par_rtrim[0] = 0x20;
			par_rtrim[1] = 0x09;
			par_rtrim_cnt = 2;
			break;

		case 'R':
			par_do_rtrim = 1;
			par_rtrim_cnt = buf_escape (par_rtrim,
				(unsigned char *) optarg[0],
				LNCONV_STR_MAX
			);
			break;

		case 's':
			par_do_lnconv = 1;
			par_eol_cnt = buf_escape (par_eol,
				(unsigned char *) optarg[0],
				LNCONV_STR_MAX
			);
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, "wb");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	r = lnconv (inp, out);

	misc_close (out);
	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
