/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/md5/md5.c                                                *
 * Created:     2000-12-10 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2000-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>
#include <lib/md5.h>

#include "md5.h"


#define MD5_BUF_SIZE MISC_BUF_SIZE


static const char *argv0 = NULL;

static int      par_enc = 0;
static int      par_dec = 0;
static int      par_bin = 0;
static int      par_pipe = 0;
static int      par_capital = 0;
static unsigned par_digits = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'b', 0, "binary", NULL, "Write the MD5 sum as binary string [no]" },
	{ 'c', 0, "capital", NULL, "Use capital letters as hex digits [no]" },
	{ 'd', 0, "decode", NULL, "Verify and remove binary MD5 sum from input [no]" },
	{ 'e', 0, "encode", NULL, "Add binary MD5 sum to output [no]" },
	{ 'i', 1, "input", "string", "Set input file name [stdin]" },
	{ 'n', 1, "digits", "int", "Group n digits in output [0]" },
	{ 'o', 1, "output", "string", "Set output file name [stdout]" },
	{ 'p', 0, "pipe", NULL, "Pass input through to stdout [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"md5: calculate MD5 message digests",
		"usage: md5 [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
int do_md5_enc (FILE *inp, FILE *out)
{
	unsigned long n;
	md5_t         md5;
	unsigned char buf[MD5_BUF_SIZE + 16];

	md5_init (&md5);

	n = fread (buf, 1, MD5_BUF_SIZE, inp);

	while (n > 0) {
		md5_calc (&md5, buf, n);

		if (fwrite (buf, 1, n, out) != n) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		n = fread (buf, 1, MD5_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	md5_done (&md5);

	md5_get_sum (&md5, buf);

	if (fwrite (buf, 1, 16, out) != 16) {
		fprintf (stderr, "%s: write error\n", argv0);
		return (1);
	}

	fflush (out);

	return (0);
}

static
int do_md5_dec (FILE *inp, FILE *out)
{
	unsigned      i;
	unsigned long n;
	md5_t         md5;
	unsigned char buf[MD5_BUF_SIZE + 32];

	md5_init (&md5);

	n = fread (buf, 1, MD5_BUF_SIZE + 16, inp);

	while (n > 16) {
		md5_calc (&md5, buf, n - 16);

		if (fwrite (buf, 1, n - 16, out) != (n - 16)) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}

		for (i = 0; i < 16; i++) {
			buf[i] = buf[n - 16 + i];
		}

		n = 16 + fread (buf + 16, 1, MD5_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	md5_done (&md5);

	md5_get_sum (&md5, buf + 16);

	for (i = 0; i < 16; i++) {
		if (buf[i] != buf[i + 16]) {
			fprintf (stderr, "%s: md5 sum mismatch\n", argv0);
			return (1);
		}
	}

	return (0);
}

static
char md5_get_hex (unsigned char val)
{
	if (val < 10) {
		return ((char) ('0' + val));
	}

	if (par_capital) {
		return ((char) ('A' + val - 10));
	}

	return ((char) ('a' + val - 10));
}

static
int do_md5 (FILE *inp, FILE *out)
{
	unsigned      i, j;
	unsigned long n;
	md5_t         md5;
	unsigned char buf[MD5_BUF_SIZE + 16];

	md5_init (&md5);

	n = fread (buf, 1, MD5_BUF_SIZE, inp);

	while (n > 0) {
		md5_calc (&md5, buf, n);

		if (par_pipe) {
			if (fwrite (buf, 1, n, stdout) != n) {
				fprintf (stderr, "%s: write error\n", argv0);
				return (1);
			}
		}

		n = fread (buf, 1, MD5_BUF_SIZE, inp);
	}

	if (ferror (inp)) {
		fprintf (stderr, "%s: read error\n", argv0);
		return (1);
	}

	md5_done (&md5);

	md5_get_sum (&md5, buf);

	if (par_bin) {
		if (fwrite (buf, 1, 16, out) != 16) {
			fprintf (stderr, "%s: write error\n", argv0);
			return (1);
		}
	}
	else {
		j = 0;
		for (i = 0; i < 16; i++) {
			fputc (md5_get_hex (buf[i] >> 4), out);
			j += 1;
			if ((par_digits > 0) && (j >= par_digits)) {
				fputc (' ', out);
				j = 0;
			}

			fputc (md5_get_hex (buf[i] & 0x0f), out);
			j += 1;
			if ((par_digits > 0) && (j >= par_digits)) {
				fputc (' ', out);
				j = 0;
			}
		}

		fputc ('\n', out);
	}

	fflush (out);

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("md5");
			return (EXIT_SUCCESS);

		case 'b':
			par_bin = 1;
			break;

		case 'c':
			par_capital = 1;
			break;

		case 'd':
			par_bin = 1;
			par_enc = 0;
			par_dec = 1;
			break;

		case 'e':
			par_bin = 1;
			par_enc = 1;
			par_dec = 0;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'n':
			if (str_get_uint (optarg[0], &par_digits)) {
				fprintf (stderr,
					"%s: bad number of digits (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_pipe = 1;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, par_bin ? "wb" : "w");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	if (par_enc) {
		r = do_md5_enc (inp, out);
	}
	else if (par_dec) {
		r = do_md5_dec (inp, out);
	}
	else {
		r = do_md5 (inp, out);
	}

	misc_close (out);
	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
