/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/mdate/mdate.c                                            *
 * Created:     2006-12-02 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2006-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/date.h>
#include <lib/getopt.h>
#include <lib/strtoint.h>

#include <time.h>
#include <sys/time.h>

#include "mdate.h"


static const char *argv0 = NULL;

static int        par_none = 1;

static int64_t    par_time = 0;

static int        par_local = 1;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "all", NULL, "Print all calendars" },
	{ 'g', 0, "gregorian", NULL, "Print gregorian calendar" },
	{ 'j', 0, "julian", NULL, "Print julian calendar" },
	{ 'J', 0, "julian-day", NULL, "Print the julian day number" },
	{ 'l', 0, "local", NULL, "Use local time as time zone" },
	{ 'M', 0, "modified-julian-day", NULL, "Print the modified julian day number" },
	{ 'p', 0, "print", NULL, "Print time in default format [default]" },
	{ 's', 1, "set-date", "string", "Set the date [current date]" },
	{ 't', 0, "time", NULL, "Print internal time representation" },
	{ 'u', 0, "utc", NULL, "Use UTC as time zone [default]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'w', 0, "day-of-week", NULL, "Print the day of week in local time" },
	{ 'x', 0, "unix", NULL, "Print unix time" },
	{ 'z', 0, "timezone", NULL, "Print the time zone" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"mdate: convert dates to various formats",
		"usage: mdate [options]",
		opts
	);

	fflush (stdout);
}

#ifdef HAVE_TIMEZONE

/*
 * get the offset between UTC and local time for a specific time in UTC
 */
static
int tm_get_utcofs_utc (int64_t *tm, int *s, unsigned *hh, unsigned *mm, unsigned *ss)
{
	time_t    t;
	int64_t   tmp;
	struct tm *tms;
	long      tz;

	*s = 1;
	*hh = 0;
	*mm = 0;
	*ss = 0;

	if (par_local == 0) {
		return (0);
	}

	/* convert to unix time */
	tmp = *tm + 946684800000LL;
	if (tmp < 0) {
		return (1);
	}
	t = (time_t) (tmp / 1000);

	tms = localtime (&t);
	if (tms == NULL) {
		return (1);
	}

	tz = timezone;

	if (tms->tm_isdst > 0) {
		/* DST is in effect. We assume the shift is one hour. */
		tz -= 3600;
	}

	if (tz <= 0) {
		*s = 1;
		*hh = -tz / 3600;
		*mm = (-tz % 3600) / 60;
		*ss = (-tz % 60);
	}
	else {
		*s = 0;
		*hh = tz / 3600;
		*mm = (tz % 3600) / 60;
		*ss = tz % 60;
	}

	return (0);
}

#else

static
int tm_get_utcofs_utc (int64_t *tm, int *s, unsigned *hh, unsigned *mm, unsigned *ss)
{
	*s = 1;
	*hh = 0;
	*mm = 0;
	*ss = 0;

	return (0);
}

#endif

/*
 * get the offset between UTC and local time for a specific time in UTC
 * additionally subtract (add if neg is set) the offset from tm
 */
static
int tm_adjust_timezone (int64_t *tm, int neg, int *s, unsigned *hh, unsigned *mm, unsigned *ss)
{
	unsigned long tz;
	int           ls;
	unsigned      lhh, lmm, lss;

	s = (s == NULL) ? &ls : s;
	hh = (hh == NULL) ? &lhh : hh;
	mm = (mm == NULL) ? &lmm : mm;
	ss = (ss == NULL) ? &lss : ss;

	if (tm_get_utcofs_utc (tm, s, hh, mm, ss)) {
		return (1);
	}

	tz = 60 * *hh + *mm;
	tz = 60 * tz + *ss;

	if ((*s == 0) != (neg == 0)) {
		*tm += 1000LL * tz;
	}
	else {
		*tm -= 1000LL * tz;
	}

	return (0);
}

static
void parse_space (const char **str)
{
	const char *s;

	s = *str;

	while (*s == ' ') {
		s += 1;
	}

	*str = s;
}

static
int parse_string (const char **str, const char *s)
{
	const char *t;

	parse_space (str);

	t = *str;

	while ((*t == *s) && (*s != 0)) {
		t += 1;
		s += 1;
	}

	if (*s != 0) {
		return (0);
	}

	*str = t;

	return (1);
}

static
int parse_int64 (const char **str, int sign, int64_t *val)
{
	const char *t;
	int        s;

	parse_space (str);

	t = *str;

	s = 0;

	if (sign) {
		if (*t == '-') {
			t += 1;
			s = 1;
		}
		else if (*t == '+') {
			t += 1;
		}
	}

	if ((*t < '0') || (*t > '9')) {
		return (0);
	}

	*val = 0;

	while ((*t >= '0') && (*t <= '9')) {
		*val = 10 * *val + (*t - '0');
		t += 1;
	}

	if (s) {
		*val = -*val;
	}

	*str = t;

	return (1);
}

static
int parse_integer (const char **str, int sign, long *val)
{
	int64_t tmp;

	if (parse_int64 (str, sign, &tmp) == 0) {
		return (0);
	}

	*val = tmp;

	return (1);
}

static
int parse_ymdhms (const char **str, int *bc,
	unsigned *yy, unsigned *mm, unsigned *dd,
	unsigned *h, unsigned *m, unsigned *s, long *tz, int *tzok)
{
	long       val;
	const char *tmp;

	tmp = *str;

	*tzok = 0;

	if (parse_integer (&tmp, 1, &val) == 0) {
		return (0);
	}
	if (val < 0) {
		*bc = 1;
		*yy = -val;
	}
	else {
		*bc = 0;
		*yy = val;
	}

	*mm = 1;
	*dd = 1;
	*h = 0;
	*m = 0;
	*s = 0;

	if (parse_string (&tmp, "-")) {
		if (parse_integer (&tmp, 0, &val) == 0) {
			return (0);
		}
		if ((val < 1) || (val > 12)) {
			return (0);
		}
		*mm = val;
	}

	if (parse_string (&tmp, "-")) {
		if (parse_integer (&tmp, 0, &val) == 0) {
			return (0);
		}
		if ((val < 1) || (val > 31)) {
			return (0);
		}
		*dd = val;
	}

	if (parse_integer (&tmp, 0, &val)) {
		if ((val < 0) || (val > 23)) {
			return (0);
		}
		*h = val;
	}

	if (parse_string (&tmp, ":")) {
		if (parse_integer (&tmp, 0, &val) == 0) {
			return (0);
		}
		if ((val < 0) || (val > 59)) {
			return (0);
		}
		*m = val;
	}
	else {
		*m = 0;
	}

	if (parse_string (&tmp, ":")) {
		if (parse_integer (&tmp, 0, &val) == 0) {
			return (0);
		}
		if ((val < 0) || (val > 59)) {
			return (0);
		}
		*s = val;
	}
	else {
		*s = 0;
	}

	if (parse_integer (&tmp, 1, &val)) {
		*tz = -3600 * val;

		if (parse_string (&tmp, ":")) {
			if (parse_integer (&tmp, 0, &val) == 0) {
				return (0);
			}

			if (*tz < 0) {
				*tz -= 60 * val;
			}
			else {
				*tz += 60 * val;
			}
		}

		*tzok = 1;
	}

	*str = tmp;

	return (1);
}

static
int parse_time (const char **str, int64_t *val)
{
	int        bc;
	unsigned   yy, mm, dd, h, m, s;
	long       tz;
	int        tzok;
	const char *tmp;

	tmp = *str;

	if (parse_string (&tmp, "J:")) {
		if (parse_ymdhms (&tmp, &bc, &yy, &mm, &dd, &h, &m, &s, &tz, &tzok) == 0) {
			return (0);
		}

		if (tm_set_julian (val, bc, yy, mm, dd)) {
			return (0);
		}

		if (tm_set_time (val, h, m, s, 0)) {
			return (0);
		}

		if (tzok) {
			*val += 1000LL * tz;
		}
		else {
			if (tm_adjust_timezone (val, 1, NULL, NULL, NULL, NULL)) {
				return (0);
			}
		}
	}
	else if (parse_string (&tmp, "U:")) {
		int64_t v;

		if (parse_int64 (&tmp, 1, &v) == 0) {
			return (0);
		}

		if (tm_set_unix (val, v)) {
			return (0);
		}
	}
	else if (parse_string (&tmp, "T:")) {
		int64_t v;

		if (parse_int64 (&tmp, 1, &v) == 0) {
			return (0);
		}

		*val = v;
	}
	else {
		parse_string (&tmp, "G:");

		if (parse_ymdhms (&tmp, &bc, &yy, &mm, &dd, &h, &m, &s, &tz, &tzok) == 0) {
			return (0);
		}

		if (tm_set_gregorian (val, bc, yy, mm, dd)) {
			return (0);
		}

		if (tm_set_time (val, h, m, s, 0)) {
			return (0);
		}

		if (tzok) {
			*val += 1000LL * tz;
		}
		else {
			if (tm_adjust_timezone (val, 1, NULL, NULL, NULL, NULL)) {
				return (0);
			}
		}
	}

	*str = tmp;

	return (1);
}

static
int tm_set_now (int64_t *tm)
{
	struct timeval tv;

	if (gettimeofday (&tv, NULL)) {
		*tm = 0;
		return (1);
	}

	*tm = (int64_t) tv.tv_sec - 946684800;
	*tm *= 1000;
	*tm += (int64_t) tv.tv_usec / 1000;

	return (0);
}

static
int tm_set_string (int64_t *tm, const char *str)
{
	int64_t tmp;

	if (parse_time (&str, &tmp) == 0) {
		return (1);
	}

	parse_space (&str);

	if (*str != 0) {
		return (1);
	}

	*tm = tmp;

	return (0);
}

static
const char *get_day_name (unsigned dow)
{
	static const char *name[7] = {
		"monday",
		"tuesday",
		"wednesday",
		"thursday",
		"friday",
		"saturday",
		"sunday"
	};

	return (name[dow % 7]);
}

static
void mdate_print (FILE *fp, int64_t tm, const char *str)
{
	int      bc;
	unsigned yy, mm, dd;
	unsigned h, m, s, ms;
	int      tzsig;
	unsigned tzh, tzm, tzs;
	unsigned wd;

	tm_adjust_timezone (&tm, 0, &tzsig, &tzh, &tzm, &tzs);
	wd = tm_get_day_of_week (tm);

	tm_get_gregorian (tm, &bc, &yy, &mm, &dd);
	tm_get_time (tm, &h, &m, &s, &ms);

	fprintf (fp, "%s%s%04u-%02u-%02u %02u:%02u:%02u %c%02u:%02u %s\n",
		str,
		bc ? "-" : "", yy, mm, dd, h, m, s,
		tzsig ? '+' : '-', tzh, tzm,
		get_day_name (wd)
	);

}

static
void mdate_print_gregorian (FILE *fp, int64_t tm, const char *str)
{
	int      bc;
	unsigned yy, mm, dd;
	unsigned h, m, s, ms;
	int      tzsig;
	unsigned tzh, tzm, tzs;

	tm_adjust_timezone (&tm, 0, &tzsig, &tzh, &tzm, &tzs);

	tm_get_gregorian (tm, &bc, &yy, &mm, &dd);
	tm_get_time (tm, &h, &m, &s, &ms);

	fprintf (fp, "%s%s%04u-%02u-%02u %02u:%02u:%02u %c%02u:%02u\n",
		str,
		bc ? "-" : "", yy, mm, dd, h, m, s,
		tzsig ? '+' : '-', tzh, tzm
	);
}

static
void mdate_print_julian (FILE *fp, int64_t tm, const char *str)
{
	int      bc;
	unsigned yy, mm, dd;
	unsigned h, m, s, ms;
	int      tzsig;
	unsigned tzh, tzm, tzs;

	tm_adjust_timezone (&tm, 0, &tzsig, &tzh, &tzm, &tzs);

	tm_get_julian (tm, &bc, &yy, &mm, &dd);
	tm_get_time (tm, &h, &m, &s, &ms);

	fprintf (fp, "%s%s%04u-%02u-%02u %02u:%02u:%02u %c%02u:%02u\n",
		str,
		bc ? "-" : "", yy, mm, dd, h, m, s,
		tzsig ? '+' : '-', tzh, tzm
	);

}

static
void mdate_print_dayofweek (FILE *fp, int64_t tm, const char *str)
{
	unsigned wd;
	int      tzsig;
	unsigned tzh, tzm, tzs;

	tm_adjust_timezone (&tm, 0, &tzsig, &tzh, &tzm, &tzs);
	wd = tm_get_day_of_week (tm);

	fprintf (fp, "%s%s\n", str, get_day_name (wd));
}

static
void mdate_print_unix (FILE *fp, int64_t tm, const char *str)
{
	int64_t u;

	tm_get_unix (tm, &u);

	fprintf (fp, "%s%lld\n", str, (long long) u);
}

static
void mdate_print_time (FILE *fp, int64_t tm, const char *str)
{
	fprintf (fp, "%s%lld\n", str, (long long) tm);
}

static
void mdate_print_jd (FILE *fp, int64_t tm, const char *str)
{
	double jd;

	tm_get_jd (tm, &jd);

	fprintf (fp, "%s%.8f\n", str, jd);
}

static
void mdate_print_mjd (FILE *fp, int64_t tm, const char *str)
{
	double jd;

	tm_get_jd (tm, &jd);

	fprintf (fp, "%s%.8f\n", str, jd - 2400000.5);
}

static
void mdate_print_timezone (FILE *fp, int64_t tm, const char *str)
{
	long     tz;
	int      tzsig;
	unsigned tzh, tzm, tzs;

	tm_adjust_timezone (&tm, 0, &tzsig, &tzh, &tzm, &tzs);

	tz = 3600 * tzh + 60 * tzm + tzs;

	if (tzsig == 0) {
		tz = -tz;
	}

	fprintf (fp, "%s%ld\n", str, tz);
}

static
void mdate_print_all (FILE *fp, int64_t tm)
{
	mdate_print_gregorian (fp, tm, "Gregorian: ");
	mdate_print_julian (fp, tm, "Julian:    ");
	mdate_print_dayofweek (fp, tm, "Day:       ");
	mdate_print_unix (fp, tm, "Unix:      ");
	mdate_print_time (fp, tm, "Time:      ");
	mdate_print_jd (fp, tm, "JD:        ");
	mdate_print_mjd (fp, tm, "MJD:       ");
}

int main (int argc, char *argv[])
{
	int  r;
	char **optarg;

	argv0 = argv[0];

	tzset();

	tm_set_now (&par_time);

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("mdate");
			return (EXIT_SUCCESS);

		case 'a':
			par_none = 0;
			mdate_print_all (stdout, par_time);
			break;

		case 'g':
			par_none = 0;
			mdate_print_gregorian (stdout, par_time, "");
			break;

		case 'j':
			par_none = 0;
			mdate_print_julian (stdout, par_time, "");
			break;

		case 'J':
			par_none = 0;
			mdate_print_jd (stdout, par_time, "");
			break;

		case 'l':
			par_local = 1;
			break;

		case 'M':
			par_none = 0;
			mdate_print_mjd (stdout, par_time, "");
			break;

		case 'p':
			par_none = 0;
			mdate_print (stdout, par_time, "");
			break;

		case 's':
			if (tm_set_string (&par_time, optarg[0])) {
				fprintf (stderr, "%s: bad date (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 't':
			par_none = 0;
			mdate_print_time (stdout, par_time, "");
			break;

		case 'u':
			par_local = 0;
			break;

		case 'w':
			par_none = 0;
			mdate_print_dayofweek (stdout, par_time, "");
			break;

		case 'x':
			par_none = 0;
			mdate_print_unix (stdout, par_time, "");
			break;

		case 'z':
			par_none = 0;
			mdate_print_timezone (stdout, par_time, "");
			break;

		case 0:
			if (tm_set_string (&par_time, optarg[0])) {
				fprintf (stderr, "%s: bad date (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_none) {
		mdate_print (stdout, par_time, "");
	}

	return (EXIT_SUCCESS);
}
