/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/pbuf/pbuf.c                                              *
 * Created:     2005-12-15 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2005-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/strtoint.h>

#include <errno.h>
#include <unistd.h>

#include "pbuf.h"


typedef struct pbuf_s {
	struct pbuf_s *next;
	unsigned      cnt;
	unsigned char buf[16384];
} pbuf_t;


static const char    *argv0 = NULL;

static int           par_pipe = 0;
static int           par_append = 0;
static unsigned long par_max_size = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "append", NULL, "Append to the output file [no]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'm', 1, "max-size", "int", "Set the maximum buffer size [0]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'p', 0, "pipe", NULL, "Pass input through to stdout [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"pbuf: pipe buffer",
		"usage: pbuf [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
pbuf_t *pbuf_new (void)
{
	pbuf_t *buf;

	buf = malloc (sizeof (pbuf_t));
	if (buf == NULL) {
		return (NULL);
	}

	buf->next = NULL;
	buf->cnt = 0;

	return (buf);
}

static
void pbuf_del (pbuf_t *buf)
{
	pbuf_t *tmp;

	while (buf != NULL) {
		tmp = buf;
		buf = buf->next;
		free (tmp);
	}
}

static
pbuf_t *pbuf_read_buf (FILE *inp, unsigned max)
{
	unsigned i;
	ssize_t  r;
	pbuf_t   *buf;

	buf = pbuf_new();
	if (buf == NULL) {
		fprintf (stderr, "%s: memory allocation error\n", argv0);
		return (NULL);
	}

	i = 0;
	while (i < max) {
		r = read (fileno (inp), buf->buf + i, max - i);

		if (r < 0) {
			if (errno != EINTR) {
				fprintf (stderr, "%s: read error\n", argv0);
				break;
			}
		}
		else if (r == 0) {
			break;
		}
		else {
			if (par_pipe) {
				fwrite (buf->buf + i, 1, r, stdout);
				fflush (stdout);
			}

			i += r;
		}
	}

	buf->cnt = i;

	if (i == 0) {
		pbuf_del (buf);
		return (NULL);
	}

	return (buf);
}

static
void pbuf_skip (FILE *inp)
{
	pbuf_t *buf;

	while (1) {
		buf = pbuf_read_buf (inp, sizeof (buf->buf));

		if (buf == NULL) {
			return;
		}

		pbuf_del (buf);
	}
}

static
pbuf_t *pbuf_read (FILE *inp)
{
	unsigned      n;
	unsigned long total;
	pbuf_t        *head, *tail, *buf;

	head = NULL;
	tail = NULL;

	total = 0;

	while (1) {
		n = sizeof (buf->buf);

		if (par_max_size > 0) {
			unsigned long m;

			if (total >= par_max_size) {
				pbuf_skip (inp);
				break;
			}

			m = (par_max_size - total);
			if (m < n) {
				n = m;
			}
		}

		/* n > 0 */

		buf = pbuf_read_buf (inp, n);
		if (buf == NULL) {
			break;
		}

		total += buf->cnt;

		if (head == NULL) {
			head = buf;
		}
		else {
			tail->next = buf;
		}

		tail = buf;
	}

	if (head == NULL) {
		head = pbuf_new();
	}

	return (head);
}

static
int pbuf_write (FILE *out, pbuf_t *buf)
{
	pbuf_t *tmp;

	while (buf != NULL) {
		tmp = buf;
		buf = buf->next;

		if (fwrite (tmp->buf, 1, tmp->cnt, out) != tmp->cnt) {
			fprintf (stderr, "%s: write error\n", argv0);
			pbuf_del (tmp);
			return (1);
		}

		free (tmp);
	}

	return (0);
}

static
int pbuf (const char *finp, const char *fout)
{
	FILE   *fp;
	pbuf_t *buf;

	if ((finp == NULL) || (strcmp (finp, "-") == 0)) {
		fp = stdin;
	}
	else {
		fp = fopen (finp, "rb");
		if (fp == NULL) {
			fprintf (stderr, "%s: can't open input file (%s)\n",
				argv0, finp
			);
			return (1);
		}
	}

	buf = pbuf_read (fp);
	if (buf == NULL) {
		return (1);
	}

	if (fp != stdin) {
		fclose (fp);
	}

	if ((fout == NULL) || (strcmp (fout, "-") == 0)) {
		fp = stdout;
	}
	else {
		fp = fopen (fout, par_append ? "ab" : "wb");
		if (fp == NULL) {
			fprintf (stderr, "%s: can't open output file (%s)\n",
				argv0, fout
			);
			pbuf_del (buf);
			return (1);
		}
	}

	if (pbuf_write (fp, buf)) {
		return (1);
	}

	if (fp != stdout) {
		fclose (fp);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;

	argv0 = argv[0];

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("pbuf");
			return (EXIT_SUCCESS);

		case 'a':
			par_append = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'm':
			if (str_get_ulng (optarg[0], &par_max_size)) {
				fprintf (stderr, "%s: bad maximum size (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_pipe = 1;
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if (pbuf (finp, fout)) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
