/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/quote/quote.c                                            *
 * Created:     2004-05-21 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2004-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>
#include <lib/strtoint.h>

#include "quote.h"


static const char    *argv0 = NULL;

static unsigned      par_enc = 1;
static unsigned      par_quote = '=';
static unsigned char par_chars[256];


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 1, "quote-char", "char", "Quote a character" },
	{ 'A', 2, "quote-range", "char char", "Quote a range of characters" },
	{ 'd', 0, "decode", NULL, "Unquote" },
	{ 'e', 0, "encode", NULL, "Quote [default]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'p', 0, "quoted-printable", NULL, "Set defaults for quoted-printable [no]" },
	{ 'q', 1, "quote", "char", "Set the quote character [=]" },
	{ 'r', 1, "no-quote-char", "char", "Don't quote a character" },
	{ 'R', 2, "no-quote-range", "char char", "Don't quote a range of characters" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"quote: quote and unquote text",
		"usage: quote [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
void put_hex (FILE *fp, unsigned val)
{
	if (val < 10) {
		fputc ('0' + val, fp);
	}
	else {
		fputc ('a' + (val - 10), fp);
	}
}

static
int get_hex (FILE *fp, unsigned *val)
{
	unsigned i;
	int      c;

	*val = 0;

	for (i = 0; i < 2; i++) {
		c = fgetc (fp);
		if ((c >= '0') && (c <= '9')) {
			*val = (*val << 4) + (c - '0');
		}
		else if ((c >= 'a') && (c <= 'f')) {
			*val = (*val << 4) + (c - 'a' + 10);
		}
		else if ((c >= 'A') && (c <= 'F')) {
			*val = (*val << 4) + (c - 'A' + 10);
		}
		else {
			return (1);
		}
	}

	return (0);
}

static
int get_char (const char *str, unsigned *val)
{
	if ((str[0] >= '0') && (str[0] <= '9')) {
		if (str_get_uint (str, val)) {
			fprintf (stderr, "%s: bad character (%s)\n",
				argv0, str
			);
			return (1);
		}
	}
	else {
		*val = str[0];
	}

	return (0);
}

static
void set_default_quote (void)
{
	unsigned i;

	for (i = 0; i < 32; i++) {
		par_chars[i] = 1;
	}

	for (i = 32; i < 128; i++) {
		par_chars[i] = 0;
	}

	for (i = 128; i < 256; i++) {
		par_chars[i] = 1;
	}

	par_chars[0x09] = 0;
	par_chars[0x0a] = 0;
	par_chars[0x0d] = 0;

	par_quote = '=';
}

static
void set_quoted_printable (void)
{
	unsigned i;

	for (i = 0; i < 32; i++) {
		par_chars[i] = 1;
	}

	for (i = 32; i < 128; i++) {
		par_chars[i] = 0;
	}

	for (i = 128; i < 256; i++) {
		par_chars[i] = 1;
	}

	par_chars[0x0a] = 0;
	par_chars[0x0d] = 0;
	par_chars[0x20] = 1;

	par_quote = '=';
}

static
int quote_enc (FILE *inp, FILE *out, unsigned char q)
{
	int c;

	c = fgetc (inp);
	while (c != EOF) {
		if ((par_chars[c] != 0) || (c == q)) {
			fputc (q, out);
			put_hex (out, (c >> 4) & 0x0f);
			put_hex (out, c & 0x0f);
		}
		else {
			fputc (c, out);
		}

		c = fgetc (inp);
	}

	return (0);
}

static
int quote_dec (FILE *inp, FILE *out, unsigned char q)
{
	int      c;
	unsigned val;

	c = fgetc (inp);
	while (c != EOF) {
		if (c == q) {
			if (get_hex (inp, &val)) {
				return (1);
			}

			fputc (val, out);
		}
		else {
			fputc (c, out);
		}

		c = fgetc (inp);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	FILE       *inp, *out;
	unsigned   c1, c2;

	argv0 = argv[0];

	set_default_quote();

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("quote");
			return (EXIT_SUCCESS);

		case 'a':
			if (get_char (optarg[0], &c1)) {
				return (EXIT_FAILURE);
			}
			par_chars[c1 & 0xff] = 1;
			break;

		case 'A':
			if (get_char (optarg[0], &c1) || get_char (optarg[1], &c2)) {
				return (EXIT_FAILURE);
			}

			c1 &= 0xff;
			c2 &= 0xff;

			while (c1 <= c2) {
				par_chars[c1++] = 1;
			}
			break;

		case 'd':
			par_enc = 0;
			break;

		case 'e':
			par_enc = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			set_quoted_printable();
			break;

		case 'q':
			if (get_char (optarg[0], &par_quote)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'r':
			if (get_char (optarg[0], &c1)) {
				return (EXIT_FAILURE);
			}
			par_chars[c1 & 0xff] = 0;
			break;

		case 'R':
			if (get_char (optarg[0], &c1) || get_char (optarg[1], &c2)) {
				return (EXIT_FAILURE);
			}

			c1 &= 0xff;
			c2 &= 0xff;

			while (c1 <= c2) {
				par_chars[c1++] = 0;
			}
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	out = misc_open_out (argv0, fout, "wb");

	if (out == NULL) {
		return (EXIT_FAILURE);
	}

	if (par_enc) {
		r = quote_enc (inp, out, par_quote);
	}
	else {
		r = quote_dec (inp, out, par_quote);
	}

	misc_close (out);
	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
