/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/rpipe/rpipe.c                                            *
 * Created:     2002-01-17 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/socket.h>
#include <lib/strtoint.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "rpipe.h"


#define RPIPE_BUF_SIZE MISC_BUF_SIZE


static const char     *argv0 = NULL;

static int            par_verbose = 0;
static char           *par_host = NULL;
static char           *par_addr = NULL;
static unsigned short par_port = 0;
static int            par_input = 0;
static int            par_term = 0;
static unsigned       par_retry_delay = 2;
static unsigned long  par_retry_cnt = 0;

static int            par_reading = 0;
static int            par_writing = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'h', 1, "host", "string[:int]", "Set the host name and port" },
	{ 'i', 0, "input", NULL, "Listen to a port (default if -p is used)" },
	{ 'n', 1, "retry-count", "int", "Set the retry count [0]" },
	{ 'o', 0, "output", NULL, "Connect to a port (default if -h is used)" },
	{ 'p', 1, "port", "int", "Set the port" },
	{ 'r', 0, "read", NULL, "Read from pipe (default if input)" },
	{ 't', 1, "retry-delay", "int", "Set the retry delay in seconds [2]" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'w', 0, "write", NULL, "Write to pipe (default if output)" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"rpipe: extend pipes over a network",
		"usage: rpipe [options]",
		opts
	);

	fflush (stdout);
}

static
void sig_term (int sig)
{
	par_term = 1;
}

static
void sig_pipe (int sig)
{
}

static
int rpipe_copy (int src, int dst)
{
	ssize_t       cnt;
	unsigned      buf_i, buf_n;
	unsigned char buf[RPIPE_BUF_SIZE];


	cnt = read (src, buf, RPIPE_BUF_SIZE);
	if (par_term || (cnt < 0)) {
		return (1);
	}

	while (cnt > 0) {
		buf_i = 0;
		buf_n = (unsigned) cnt;

		while (buf_n > 0) {
			cnt = write (dst, buf + buf_i, buf_n);
			if (par_term || (cnt <= 0)) {
				return (1);
			}

			buf_i += (unsigned) cnt;
			buf_n -= (unsigned) cnt;
		}

		cnt = read (src, buf, RPIPE_BUF_SIZE);
		if (par_term || (cnt < 0)) {
			return (1);
		}
	}

	return (0);
}

static
int rpipe_handle (int fd)
{
	int   r;
	int   status;
	pid_t pid, tmp;

	if (par_reading && par_writing) {
		pid = fork();
		if (pid < 0) {
			fprintf (stderr, "%s: fork failed\n", argv0);
			return (1);
		}

		if (pid == 0) {
			r = rpipe_copy (fd, 1);
			tcp_shutdown_inp (fd);
			exit (r);
		}

		r = rpipe_copy (0, fd);
		tcp_shutdown_out (fd);

		while (par_term == 0) {
			tmp = waitpid (pid, &status, 0);
			if (tmp < 0) {
				return (1);
			}

			if (tmp == pid) {
				return (r || status);
			}
		}

		return (r);
	}

	if (par_reading) {
		return (rpipe_copy (fd, 1));
	}

	if (par_writing) {
		return (rpipe_copy (0, fd));
	}

	return (1);
}

static
int rpipe_inp (void)
{
	int r;
	int fd1, fd2;

	fd1 = tcp_listen (par_addr, par_port, 16);
	if (fd1 == -1) {
		fprintf (stderr, "%s: can't create socket for port %u\n", argv0, par_port);
		return (1);
	}

	fd2 = -1;

	while ((fd2 < 0) && (par_term == 0)) {
		fd2 = tcp_accept_time (fd1, 500);
	}

	close (fd1);

	if (fd2 < 0) {
		return (1);
	}

	if (par_verbose) {
		char           *host;
		unsigned short port;

		if (tcp_get_peer (fd2, &host, &port) == 0) {
			fprintf (stderr, "rpipe: accepted connection from %s:%u\n", host, port);
			fflush (stderr);
			free (host);
		}
	}

	r = rpipe_handle (fd2);

	close (fd2);

	return (r);
}

static
int rpipe_out (void)
{
	int r;
	int fd;

	if (par_addr == NULL) {
		fprintf (stderr, "%s: need a host name\n", argv0);
		return (1);
	}

	fd = -1;

	while (fd < 0) {
		if (par_term) {
			return (1);
		}

		fd = tcp_connect (par_addr, par_port);

		if (fd < 0) {
			unsigned tm;

			if (par_retry_cnt > 0) {
				par_retry_cnt -= 1;

				if (par_retry_cnt == 0) {
					fprintf (stderr, "%s: exceeded retry count\n", argv0);
					return (1);
				}
			}

			tm = par_retry_delay;
			while (tm > 0) {
				if (par_term) {
					return (1);
				}

				tm = sleep (tm);
			}
		}
	}

	if (par_verbose) {
		fprintf (stderr, "rpipe: connected to %s:%u\n", par_addr, (unsigned) par_port);
		fflush (stderr);
	}

	r = rpipe_handle (fd);

	close (fd);

	return (r);
}

int main (int argc, char *argv[])
{
	int            r;
	char           **optarg;
	unsigned short port;
	unsigned       tmp;
	int            have_dir;

	argv0 = argv[0];

	have_dir = 0;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("rpipe");
			return (EXIT_SUCCESS);

		case 'h':
			par_host = tcp_addr_get_host (optarg[0]);

			if (tcp_addr_get_port (optarg[0], &port) == 0) {
				par_port = port;
			}

			if (!have_dir) {
				par_input = 0;
				have_dir = 1;
			}
			break;

		case 'i':
			par_input = 1;
			have_dir = 1;
			break;

		case 'n':
			if (str_get_ulng (optarg[0], &par_retry_cnt)) {
				fprintf (stderr, "%s: bad retry count (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			par_input = 0;
			have_dir = 1;
			break;

		case 'p':
			if (str_get_uint (optarg[0], &tmp)) {
				fprintf (stderr, "%s: bad port number (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}

			par_port = tmp;

			if (!have_dir) {
				par_input = 1;
				have_dir = 1;
			}
			break;

		case 'r':
			par_reading = 1;
			break;

		case 't':
			if (str_get_uint (optarg[0], &par_retry_delay)) {
				fprintf (stderr, "%s: bad retry delay (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'v':
			par_verbose = 1;
			break;

		case 'w':
			par_writing = 1;
			break;

		case 0:
			fprintf (stderr, "%s: unknown parameter (%s)\n",
				argv0, optarg[0]
			);
			return (EXIT_FAILURE);

		default:
			return (EXIT_FAILURE);
		}
	}

	if (par_host != NULL) {
		par_addr = tcp_host_resolve (par_host);
	}

	if ((par_reading == 0) && (par_writing == 0)) {
		if (par_input) {
			par_reading = 1;
		}
		else {
			par_writing = 1;
		}
	}

	signal (SIGINT, &sig_term);
	signal (SIGTERM, &sig_term);
	signal (SIGPIPE, &sig_pipe);

	if (par_input) {
		r = rpipe_inp();
	}
	else {
		r = rpipe_out();
	}

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
