/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/sleep/sleep.c                                            *
 * Created:     2001-12-11 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2001-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "sleep.h"


static const char         *argv0 = NULL;

static unsigned long long par_time = 0;
static int                par_verbose = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'v', 0, "verbose", NULL, "Verbose operation [no]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"sleep: sleep for a specified period of time",
		"usage: sleep [options] [time...]",
		opts
	);

	fputs ("\n"
		"Time is specified in seconds except if one of these suffixes is used:\n"
		"  d   days\n"
		"  h   hours\n"
		"  m   minutes\n"
		"  s   seconds\n"
		"  ms  milliseconds\n"
		"  u   microseconds\n",
		stdout
	);

	fflush (stdout);
}

static
int is_digit (char c)
{
	if ((c >= '0') && (c <= '9')) {
		return (1);
	}

	return (0);
}

static
int time_from_str (unsigned long long *tm, const char *str)
{
	unsigned           i;
	unsigned           dig;
	unsigned long long val;

	if (strchr (str, '.') != NULL) {
		double val;
		char   *end;

		val = strtod (str, &end);
		if ((end != NULL) && (*end != 0)) {
			return (1);
		}

		*tm = (unsigned long long) (1.0e6 * val);

		return (0);
	}

	val = 0;

	i = 0;
	while (is_digit (str[i])) {
		dig = (unsigned) (str[i] - '0');

		val = 10 * val + dig;

		i += 1;
	}

	if (str[i] == 0) {
		*tm = 1000000 * val;
		return (0);
	}

	switch (str[i]) {
	case 'd':
		val *= 24 * 60 * 60;
		val *= 1000000;
		break;

	case 'h':
		val *= 60 * 60;
		val *= 1000000;
		break;

	case 'm':
		if (str[i + 1] == 's') {
			i += 1;
			val *= 1000;
		}
		else {
			val *= 60;
			val *= 1000000;
		}
		break;

	case 'u':
		if (str[i + 1] == 's') {
			i += 1;
		}
		break;

	case 's':
		val *= 1000000;
		break;

	default:
		return (1);
	}

	*tm = val;

	if (str[i + 1] != 0) {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int                r;
	char               **optarg;
	unsigned long long val;
	unsigned long      sleep_s;
	unsigned long      sleep_u;
	struct timeval     tv;

	argv0 = argv[0];

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("sleep");
			return (EXIT_SUCCESS);

		case 'v':
			par_verbose = 1;
			break;

		case 0:
			if (time_from_str (&val, optarg[0])) {
				fprintf (stderr, "%s: bad time format (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			par_time += val;
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	sleep_u = (unsigned long) (par_time % 1000000);
	sleep_s = (unsigned long) (par_time / 1000000);

	if (par_verbose) {
		fprintf (stdout, "sleeping %lu seconds + %lu microseconds\n", sleep_s, sleep_u);
	}

	while (sleep_s > 0) {
		sleep_s = sleep ((unsigned) sleep_s);
	}

	if (sleep_u > 0) {
		tv.tv_sec = 0;
		tv.tv_usec = (long) sleep_u;

		select (0, NULL, NULL, NULL, &tv);
	}

	return (EXIT_SUCCESS);
}
