/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/tjoint/tjoint.c                                          *
 * Created:     1999-03-17 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 1999-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/getopt.h>
#include <lib/open.h>

#include "tjoint.h"


static const char *argv0 = NULL;

static unsigned   par_file_cnt = 0;
static FILE       **par_file = NULL;

static int        par_append = 0;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'a', 0, "append", NULL, "Append to following output files [no]" },
	{ 'b', 1, "buffer", "int", "Set the buffer size" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'o', 1, "output", "string", "Set an output file name [none]" },
	{ 'O', 1, "output-append", "string", "Set an output file name and append [none]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{ 'w', 0, "overwrite", NULL, "Overwrite the following files [default]" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"tjoint: split output to multiple files",
		"usage: tjoint [options] [filename...]",
		opts
	);

	fflush (stdout);
}

static
FILE **file_new (void)
{
	FILE **tmp;

	tmp = realloc (par_file, (par_file_cnt + 1) * sizeof (FILE *));

	if (tmp == NULL) {
		return (NULL);
	}

	tmp[par_file_cnt] = NULL;

	par_file_cnt += 1;
	par_file = tmp;

	return (par_file + par_file_cnt - 1);
}

static
int tjoint_new_fname (const char *fname, int append)
{
	FILE **fp;

	fp = file_new();

	if (fp == NULL) {
		return (1);
	}

	append = (par_append || append);

	*fp = misc_open_out (argv0, fname, append ? "ab" : "wb");

	if (*fp == NULL) {
		fprintf (stderr, "%s: can't open file (%s)\n", argv0, fname);
		return (1);
	}

	return (0);
}

static
int do_tjoint (FILE *inp)
{
	unsigned i;
	int      c;

	while (1) {
		c = fgetc (inp);

		if (c == EOF) {
			break;
		}

		for (i = 0; i < par_file_cnt; i++) {
			fputc (c, par_file[i]);
		}
	}

	for (i = 0; i < par_file_cnt; i++) {
		fflush (par_file[i]);
		misc_close (par_file[i]);
	}

	if (ferror (inp)) {
		return (1);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp;
	FILE       *inp;

	argv0 = argv[0];

	finp = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("tjoint");
			return (EXIT_SUCCESS);

		case 'a':
			par_append = 1;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'o':
			if (tjoint_new_fname (optarg[0], 0)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'O':
			if (tjoint_new_fname (optarg[0], 1)) {
				return (EXIT_FAILURE);
			}
			break;

		case 'w':
			par_append = 0;
			break;

		case 0:
			if (tjoint_new_fname (optarg[0], 0)) {
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	inp = misc_open_inp (argv0, finp, "rb");

	if (inp == NULL) {
		return (EXIT_FAILURE);
	}

	r = do_tjoint (inp);

	misc_close (inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
