/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:   src/ttylog/ttylog.c                                          *
 * Created:     2002-02-04 by Hampa Hug <hampa@hampa.ch>                     *
 * Copyright:   (C) 2002-2009 Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/


#include <lib/common.h>
#include <lib/file.h>
#include <lib/getopt.h>
#include <lib/strtoint.h>

#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "ttylog.h"


static const char    *argv0 = NULL;

static int           par_decode = 0;
static int           par_pipe = 0;

static unsigned long par_scale = 100;
static unsigned long par_max = 5000;
static unsigned long par_threshold = 2;

static file_t        par_fpipe;


static misc_option_t opts[] = {
	{ '?', 0, "help", NULL, "Print usage information" },
	{ 'd', 0, "decode", NULL, "Decode a character stream" },
	{ 'e', 0, "encode", NULL, "Encode a character stream [default]" },
	{ 'i', 1, "input", "string", "Set the input file name [stdin]" },
	{ 'm', 1, "max", "int", "Set the maximum delay in milliseconds [5000]" },
	{ 'o', 1, "output", "string", "Set the output file name [stdout]" },
	{ 'p', 0, "pipe", NULL, "Copy input unchanged to stdout [no]" },
	{ 's', 1, "scale", "int", "Set time scale in percent [100]" },
	{ 't', 1, "threshold", "int", "Set the threshold delay in milliseconds [2]" },
	{ 'V', 0, "version", NULL, "Print version information" },
	{  -1, 0, NULL, NULL, NULL }
};


static
void print_help (void)
{
	misc_getopt_help (
		"ttylog: log a character stream with timing information",
		"usage: ttylog [options] [input [output]]",
		opts
	);

	fflush (stdout);
}

static
unsigned long get_delta_time (struct timeval *tv)
{
	unsigned long  ret;
	long           tmp;
	struct timeval tvnew;

	gettimeofday (&tvnew, NULL);

	ret = 1000 * (unsigned long) (tvnew.tv_sec - tv->tv_sec);
	tmp = (tvnew.tv_usec - tv->tv_usec) / 1000;

	if (tmp < 0) {
		ret -= 1000;
		tmp += 1000;
	}

	ret += (unsigned long) tmp;

	if (ret <= par_threshold) {
		ret = 0;
	}

	*tv = tvnew;

	return (ret);
}

static
void set_delta_time (unsigned long tm)
{
	struct timeval tv;

	tm = (par_scale * tm) / 100;

	if (tm > par_max) {
		tm = par_max;
	}

	if (tm > 0) {
		tv.tv_sec = (long) (tm / 1000);
		tv.tv_usec = (long) (1000 * (tm % 1000));

		select (0, NULL, NULL, NULL, &tv);
	}
}

static
int do_encode (file_t *f_inp, file_t *f_out)
{
	int            c;
	unsigned long  tm;
	unsigned char  buf[5];
	struct timeval tv;

	gettimeofday (&tv, NULL);
	c = file_get_char (f_inp);

	while (c != EOF) {
		tm = get_delta_time (&tv);

		buf[0] = (tm >> 24) & 0xff;
		buf[1] = (tm >> 16) & 0xff;
		buf[2] = (tm >> 8) & 0xff;
		buf[3] = tm & 0xff;
		buf[4] = c & 0xff;

		file_write_block (f_out, buf, 5);
		file_flush (f_out);

		if (par_pipe) {
			file_set_char (&par_fpipe, c);
			file_flush (&par_fpipe);
		}

		c = file_get_char (f_inp);
	}

	return (0);
}

static
int do_decode (file_t *f_inp, file_t *f_out)
{
	int           c;
	unsigned long r;
	unsigned long tm;
	unsigned char buf[5];

	r = file_read_block (f_inp, buf, 5);

	while (r == 5) {
		tm = buf[0] & 0xff;
		tm = (tm << 8) + (buf[1] & 0xff);
		tm = (tm << 8) + (buf[2] & 0xff);
		tm = (tm << 8) + (buf[3] & 0xff);
		c = buf[4];

		if (tm > par_threshold) {
			set_delta_time (tm);
		}

		file_set_char (f_out, c);
		file_flush (f_out);

		if (par_pipe) {
			file_write_block (&par_fpipe, buf, 5);
			file_flush (&par_fpipe);
		}

		r = file_read_block (f_inp, buf, 5);
	}

	return (0);
}

int main (int argc, char *argv[])
{
	int        r;
	char       **optarg;
	const char *finp, *fout;
	file_t     inp, out;

	argv0 = argv[0];

	file_set_stdin (&inp);
	file_set_stdout (&out);

	finp = NULL;
	fout = NULL;

	while (1) {
		r = misc_getopt (argc, argv, &optarg, opts);

		if (r == GETOPT_DONE) {
			break;
		}

		if (r < 0) {
			return (EXIT_FAILURE);
		}

		switch (r) {
		case '?':
			print_help();
			return (EXIT_SUCCESS);

		case 'V':
			prt_version ("ttylog");
			return (EXIT_SUCCESS);

		case 'd':
			par_decode = 1;
			break;

		case 'e':
			par_decode = 0;
			break;

		case 'i':
			finp = optarg[0];
			break;

		case 'm':
			if (str_get_ulng (optarg[0], &par_max)) {
				fprintf (stderr,
					"%s: bad maximum delta time (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 'o':
			fout = optarg[0];
			break;

		case 'p':
			par_pipe = 1;
			file_set_stdout (&par_fpipe);
			break;

		case 's':
			if (str_get_ulng (optarg[0], &par_scale)) {
				fprintf (stderr, "%s: bad time scale (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 't':
			if (str_get_ulng (optarg[0], &par_threshold)) {
				fprintf (stderr, "%s: bad threshold (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		case 0:
			if (finp == NULL) {
				finp = optarg[0];
			}
			else if (fout == NULL) {
				fout = optarg[0];
			}
			else {
				fprintf (stderr, "%s: too many files (%s)\n",
					argv0, optarg[0]
				);
				return (EXIT_FAILURE);
			}
			break;

		default:
			return (EXIT_FAILURE);
		}
	}

	if ((finp != NULL) && (strcmp (finp, "-") != 0)) {
		file_set_name (&inp, finp);
	}

	if ((fout != NULL) && (strcmp (fout, "-") != 0)) {
		file_set_name (&out, fout);
	}

	if (file_open_inp (&inp)) {
		fprintf (stderr, "%s: can't open input file\n", argv0);
		return (EXIT_FAILURE);
	}

	if (file_open_out (&out)) {
		fprintf (stderr, "%s: can't open output file\n", argv0);
		return (EXIT_FAILURE);
	}

	if (par_decode == 0) {
		r = do_encode (&inp, &out);
	}
	else {
		r = do_decode (&inp, &out);
	}

	file_close (&out);
	file_close (&inp);

	if (r) {
		return (EXIT_FAILURE);
	}

	return (EXIT_SUCCESS);
}
